# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
from typing import Any
import unittest

from langfun.env import interface
from langfun.env import test_utils

TestingEnvironment = test_utils.TestingEnvironment
TestingSandbox = test_utils.TestingSandbox
TestingFeature = test_utils.TestingFeature
TestingEventHandler = test_utils.TestingEventHandler


class SandboxStateTests(unittest.TestCase):

  def setUp(self):
    super().setUp()
    self.event_handler = TestingEventHandler(
        log_sandbox_status=True,
        log_feature_setup=True,
        log_session_setup=True,
    )
    self.maxDiff = None

  def _create_env(
      self,
      features,
      *,
      pool_size=0,
      **kwargs
  ) -> TestingEnvironment:
    return TestingEnvironment(
        pool_size=pool_size,
        features=features,
        outage_grace_period=0,
        event_handler=self.event_handler,
        outage_retry_interval=0,
        **kwargs
    )

  def test_passive_session_setup(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(),
        },
    )
    self.assertFalse(env.enable_pooling('test_image'))
    with env:
      with env.sandbox('session1') as sb:
        sb.shell('echo "hello"')
      self.assertEqual(
          self.event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env] environment started',
              '[testing-env/test_image:0@<idle>] shell: "feature1" setup',
              '[testing-env/test_image:0/feature1] feature setup',
              '[testing-env/test_image:0@<idle>] shell: "feature2" setup',
              '[testing-env/test_image:0/feature2] feature setup',
              '[testing-env/test_image:0] created -> ready',
              '[testing-env/test_image:0] sandbox started',
              '[testing-env/test_image:0] ready -> acquired',
              '[testing-env/test_image:0] acquired -> setting_up',
              '[testing-env/test_image:0@session1] shell: "feature1" setup session',
              '[testing-env/test_image:0/feature1@session1] feature setup session',
              '[testing-env/test_image:0@session1] shell: "feature2" setup session',
              '[testing-env/test_image:0/feature2@session1] feature setup session',
              '[testing-env/test_image:0] setting_up -> in_session',
              "[testing-env/test_image:0] session 'session1' started",
              '[testing-env/test_image:0@session1] shell: echo "hello"',
              '[testing-env/test_image:0] in_session -> exiting_session',
              '[testing-env/test_image:0@session1] shell: "feature1" teardown session',
              '[testing-env/test_image:0/feature1@session1] feature teardown session',
              '[testing-env/test_image:0@session1] shell: "feature2" teardown session',
              '[testing-env/test_image:0/feature2@session1] feature teardown session',
              "[testing-env/test_image:0] session 'session1' ended",
              '[testing-env/test_image:0] exiting_session -> acquired',
              '[testing-env/test_image:0] acquired -> shutting_down',
              '[testing-env/test_image:0@<idle>] shell: "feature1" teardown',
              '[testing-env/test_image:0/feature1] feature teardown',
              '[testing-env/test_image:0@<idle>] shell: "feature2" teardown',
              '[testing-env/test_image:0/feature2] feature teardown',
              '[testing-env/test_image:0] shutting_down -> offline',
              '[testing-env/test_image:0] sandbox shutdown'
              # pylint: enable=line-too-long
          ]
      )

  def test_proactive_session_setup(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(setup_session_delay=0.1),
            'feature2': TestingFeature(),
        },
        pool_size=1,
        proactive_session_setup=True,
    )
    self.assertTrue(env.enable_pooling('test_image'))
    with env:
      with env.sandbox('session1') as sb:
        sb.shell('echo "hello"')
      sb.wait_until_not(
          (
              interface.Sandbox.Status.IN_SESSION,
              interface.Sandbox.Status.SETTING_UP
          )
      )
      self.assertEqual(sb.status, interface.Sandbox.Status.READY)
      self.assertEqual(
          self.event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env/test_image:0:0@<idle>] shell: "feature1" setup',
              '[testing-env/test_image:0:0/feature1] feature setup',
              '[testing-env/test_image:0:0@<idle>] shell: "feature2" setup',
              '[testing-env/test_image:0:0/feature2] feature setup',
              '[testing-env/test_image:0:0@<idle>] shell: "feature1" setup session',
              '[testing-env/test_image:0:0/feature1@<idle>] feature setup session',
              '[testing-env/test_image:0:0@<idle>] shell: "feature2" setup session',
              '[testing-env/test_image:0:0/feature2@<idle>] feature setup session',
              '[testing-env/test_image:0:0] created -> ready',
              '[testing-env/test_image:0:0] sandbox started',
              '[testing-env] environment started',
              '[testing-env/test_image:0:0] ready -> acquired',
              '[testing-env/test_image:0:0] acquired -> setting_up',
              '[testing-env/test_image:0:0] setting_up -> in_session',
              "[testing-env/test_image:0:0] session 'session1' started",
              '[testing-env/test_image:0:0@session1] shell: echo "hello"',
              '[testing-env/test_image:0:0] in_session -> exiting_session',
              '[testing-env/test_image:0:0@session1] shell: "feature1" teardown session',
              '[testing-env/test_image:0:0/feature1@session1] feature teardown session',
              '[testing-env/test_image:0:0@session1] shell: "feature2" teardown session',
              '[testing-env/test_image:0:0/feature2@session1] feature teardown session',
              "[testing-env/test_image:0:0] session 'session1' ended",
              '[testing-env/test_image:0:0] exiting_session -> setting_up',
              '[testing-env/test_image:0:0@<idle>] shell: "feature1" setup session',
              '[testing-env/test_image:0:0/feature1@<idle>] feature setup session',
              '[testing-env/test_image:0:0@<idle>] shell: "feature2" setup session',
              '[testing-env/test_image:0:0/feature2@<idle>] feature setup session',
              '[testing-env/test_image:0:0] setting_up -> ready'
              # pylint: enable=line-too-long
          ]
      )

  def test_proactive_session_setup_with_setup_session_error(self):
    env = self._create_env(
        features={'test_feature': TestingFeature(setup_session_delay=0.5)},
        pool_size=1,
        housekeep_interval=10.0,
    )
    with env:
      with env.sandbox('session1') as sb:
        sb.test_feature.rebind(
            simulate_setup_session_error=interface.SandboxStateError,
            skip_notification=True
        )
      sb.wait_until_not(
          (
              interface.Sandbox.Status.SETTING_UP,
              interface.Sandbox.Status.SHUTTING_DOWN
          )
      )
      self.assertEqual(len(sb.state_errors), 1)
      self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
      self.assertEqual(
          self.event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env/test_image:0:0@<idle>] shell: "test_feature" setup',
              '[testing-env/test_image:0:0/test_feature] feature setup',
              '[testing-env/test_image:0:0@<idle>] shell: "test_feature" setup session',
              '[testing-env/test_image:0:0/test_feature@<idle>] feature setup session',
              '[testing-env/test_image:0:0] created -> ready',
              '[testing-env/test_image:0:0] sandbox started',
              '[testing-env] environment started',
              '[testing-env/test_image:0:0] ready -> acquired',
              '[testing-env/test_image:0:0] acquired -> setting_up',
              '[testing-env/test_image:0:0] setting_up -> in_session',
              "[testing-env/test_image:0:0] session 'session1' started",
              '[testing-env/test_image:0:0] in_session -> exiting_session',
              '[testing-env/test_image:0:0@session1] shell: "test_feature" teardown session',
              '[testing-env/test_image:0:0/test_feature@session1] feature teardown session',
              "[testing-env/test_image:0:0] session 'session1' ended",
              '[testing-env/test_image:0:0] exiting_session -> setting_up',
              '[testing-env/test_image:0:0/test_feature@<idle>] feature setup session with SandboxStateError',
              '[testing-env/test_image:0:0] setting_up -> shutting_down',
              '[testing-env/test_image:0:0@<idle>] shell: "test_feature" teardown',
              '[testing-env/test_image:0:0/test_feature] feature teardown',
              '[testing-env/test_image:0:0] shutting_down -> offline',
              '[testing-env/test_image:0:0] sandbox shutdown'
              # pylint: enable=line-too-long
          ]
      )

  def test_sandbox_start_non_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(),
        },
        simulate_start_error=ValueError,
    )
    with env:
      with self.assertRaises(ValueError):
        with env.sandbox('session1'):
          pass
      self.assertTrue(env.is_online)
      self.assertEqual(
          self.event_handler.logs,
          [
              '[testing-env] environment started',
              '[testing-env/test_image:0] sandbox started with ValueError',
              '[testing-env/test_image:0] created -> shutting_down',
              '[testing-env/test_image:0] shutting_down -> offline',
              '[testing-env/test_image:0] sandbox shutdown'
          ]
      )

  def test_sandbox_start_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(),
        },
        pool_size=1,
        simulate_start_error=interface.SandboxStateError,
    )
    with self.assertRaises(interface.EnvironmentOutageError):
      with env:
        pass
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env/test_image:0:0] sandbox started with SandboxStateError',
            '[testing-env/test_image:0:0] created -> shutting_down',
            '[testing-env/test_image:0:0] shutting_down -> offline',
            '[testing-env/test_image:0:0] sandbox shutdown',
            '[testing-env] environment started with EnvironmentOutageError',
            '[testing-env] environment shutdown'
            # pylint: enable=line-too-long
        ]
    )

  def test_sandbox_shutdown_non_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(),
        },
        simulate_shutdown_error=ValueError,
    )
    with env:
      with self.assertRaises(ValueError):
        with env.sandbox('session1') as sb:
          sb.shell('echo "hello"')
      self.assertEqual(len(sb.state_errors), 0)
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/test_image:0@<idle>] shell: "feature1" setup',
            '[testing-env/test_image:0/feature1] feature setup',
            '[testing-env/test_image:0@<idle>] shell: "feature2" setup',
            '[testing-env/test_image:0/feature2] feature setup',
            '[testing-env/test_image:0] created -> ready',
            '[testing-env/test_image:0] sandbox started',
            '[testing-env/test_image:0] ready -> acquired',
            '[testing-env/test_image:0] acquired -> setting_up',
            '[testing-env/test_image:0@session1] shell: "feature1" setup session',
            '[testing-env/test_image:0/feature1@session1] feature setup session',
            '[testing-env/test_image:0@session1] shell: "feature2" setup session',
            '[testing-env/test_image:0/feature2@session1] feature setup session',
            '[testing-env/test_image:0] setting_up -> in_session',
            "[testing-env/test_image:0] session 'session1' started",
            '[testing-env/test_image:0@session1] shell: echo "hello"',
            '[testing-env/test_image:0] in_session -> exiting_session',
            '[testing-env/test_image:0@session1] shell: "feature1" teardown session',
            '[testing-env/test_image:0/feature1@session1] feature teardown session',
            '[testing-env/test_image:0@session1] shell: "feature2" teardown session',
            '[testing-env/test_image:0/feature2@session1] feature teardown session',
            "[testing-env/test_image:0] session 'session1' ended",
            '[testing-env/test_image:0] exiting_session -> acquired',
            '[testing-env/test_image:0] acquired -> shutting_down',
            '[testing-env/test_image:0@<idle>] shell: "feature1" teardown',
            '[testing-env/test_image:0/feature1] feature teardown',
            '[testing-env/test_image:0@<idle>] shell: "feature2" teardown',
            '[testing-env/test_image:0/feature2] feature teardown',
            '[testing-env/test_image:0] shutting_down -> offline',
            '[testing-env/test_image:0] sandbox shutdown with ValueError',
            '[testing-env] environment shutdown',
            # pylint: enable=line-too-long
        ]
    )

  def test_env_shutdown_non_state_error(self):
    env = self._create_env(
        pool_size=1,
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(),
        },
        simulate_shutdown_error=ValueError,
    )
    with self.assertRaises(ValueError):
      with env:
        pass

    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env/test_image:0:0@<idle>] shell: "feature1" setup',
            '[testing-env/test_image:0:0/feature1] feature setup',
            '[testing-env/test_image:0:0@<idle>] shell: "feature2" setup',
            '[testing-env/test_image:0:0/feature2] feature setup',
            '[testing-env/test_image:0:0@<idle>] shell: "feature1" setup session',
            '[testing-env/test_image:0:0/feature1@<idle>] feature setup session',
            '[testing-env/test_image:0:0@<idle>] shell: "feature2" setup session',
            '[testing-env/test_image:0:0/feature2@<idle>] feature setup session',
            '[testing-env/test_image:0:0] created -> ready',
            '[testing-env/test_image:0:0] sandbox started',
            '[testing-env] environment started',
            '[testing-env/test_image:0:0] ready -> shutting_down',
            '[testing-env/test_image:0:0@<idle>] shell: "feature1" teardown',
            '[testing-env/test_image:0:0/feature1] feature teardown',
            '[testing-env/test_image:0:0@<idle>] shell: "feature2" teardown',
            '[testing-env/test_image:0:0/feature2] feature teardown',
            '[testing-env/test_image:0:0] shutting_down -> offline',
            '[testing-env/test_image:0:0] sandbox shutdown with ValueError',
            '[testing-env] environment shutdown with ValueError'
            # pylint: enable=line-too-long
        ]
    )

  def test_sandbox_shutdown_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(),
        },
        simulate_shutdown_error=interface.SandboxStateError,
    )
    with env:
      with env.sandbox('session1') as sb:
        sb.shell('echo "hello"')
      self.assertEqual(len(sb.state_errors), 1)

    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/test_image:0@<idle>] shell: "feature1" setup',
            '[testing-env/test_image:0/feature1] feature setup',
            '[testing-env/test_image:0@<idle>] shell: "feature2" setup',
            '[testing-env/test_image:0/feature2] feature setup',
            '[testing-env/test_image:0] created -> ready',
            '[testing-env/test_image:0] sandbox started',
            '[testing-env/test_image:0] ready -> acquired',
            '[testing-env/test_image:0] acquired -> setting_up',
            '[testing-env/test_image:0@session1] shell: "feature1" setup session',
            '[testing-env/test_image:0/feature1@session1] feature setup session',
            '[testing-env/test_image:0@session1] shell: "feature2" setup session',
            '[testing-env/test_image:0/feature2@session1] feature setup session',
            '[testing-env/test_image:0] setting_up -> in_session',
            "[testing-env/test_image:0] session 'session1' started",
            '[testing-env/test_image:0@session1] shell: echo "hello"',
            '[testing-env/test_image:0] in_session -> exiting_session',
            '[testing-env/test_image:0@session1] shell: "feature1" teardown session',
            '[testing-env/test_image:0/feature1@session1] feature teardown session',
            '[testing-env/test_image:0@session1] shell: "feature2" teardown session',
            '[testing-env/test_image:0/feature2@session1] feature teardown session',
            "[testing-env/test_image:0] session 'session1' ended",
            '[testing-env/test_image:0] exiting_session -> acquired',
            '[testing-env/test_image:0] acquired -> shutting_down',
            '[testing-env/test_image:0@<idle>] shell: "feature1" teardown',
            '[testing-env/test_image:0/feature1] feature teardown',
            '[testing-env/test_image:0@<idle>] shell: "feature2" teardown',
            '[testing-env/test_image:0/feature2] feature teardown',
            '[testing-env/test_image:0] shutting_down -> offline',
            '[testing-env/test_image:0] sandbox shutdown with SandboxStateError',
            '[testing-env] environment shutdown',
            # pylint: enable=line-too-long
        ]
    )

  def test_feature_setup_non_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(
                simulate_setup_error=ValueError
            ),
        },
    )
    with env:
      with self.assertRaises(ValueError):
        with env.sandbox('session1'):
          pass
      self.assertEqual(
          self.event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env] environment started',
              '[testing-env/test_image:0@<idle>] shell: "feature1" setup',
              '[testing-env/test_image:0/feature1] feature setup',
              '[testing-env/test_image:0/feature2] feature setup with ValueError',
              '[testing-env/test_image:0] sandbox started with ValueError',
              '[testing-env/test_image:0] created -> shutting_down',
              '[testing-env/test_image:0@<idle>] shell: "feature1" teardown',
              '[testing-env/test_image:0/feature1] feature teardown',
              '[testing-env/test_image:0@<idle>] shell: "feature2" teardown',
              '[testing-env/test_image:0/feature2] feature teardown',
              '[testing-env/test_image:0] shutting_down -> offline',
              '[testing-env/test_image:0] sandbox shutdown'
              # pylint: enable=line-too-long
          ]
      )

  def test_feature_setup_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(
                simulate_setup_error=interface.SandboxStateError
            ),
            'feature2': TestingFeature(),
        },
    )
    with env:
      with self.assertRaises(interface.EnvironmentOutageError):
        with env.sandbox('session1'):
          pass
      self.assertEqual(
          self.event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env] environment started',
              '[testing-env/test_image:0/feature1] feature setup with SandboxStateError',
              '[testing-env/test_image:0] sandbox started with SandboxStateError',
              '[testing-env/test_image:0] created -> shutting_down',
              '[testing-env/test_image:0@<idle>] shell: "feature1" teardown',
              '[testing-env/test_image:0/feature1] feature teardown',
              '[testing-env/test_image:0] shutting_down -> offline',
              '[testing-env/test_image:0] sandbox shutdown',
              '[testing-env] environment shutdown',
              # pylint: enable=line-too-long
          ]
      )

  def test_feature_teardown_non_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(
                simulate_teardown_error=ValueError
            ),
        },
    )
    with env:
      with self.assertRaises(interface.FeatureTeardownError):
        with env.sandbox('session1'):
          pass
      self.assertEqual(
          self.event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env] environment started',
              '[testing-env/test_image:0@<idle>] shell: "feature1" setup',
              '[testing-env/test_image:0/feature1] feature setup',
              '[testing-env/test_image:0@<idle>] shell: "feature2" setup',
              '[testing-env/test_image:0/feature2] feature setup',
              '[testing-env/test_image:0] created -> ready',
              '[testing-env/test_image:0] sandbox started',
              '[testing-env/test_image:0] ready -> acquired',
              '[testing-env/test_image:0] acquired -> setting_up',
              '[testing-env/test_image:0@session1] shell: "feature1" setup session',
              '[testing-env/test_image:0/feature1@session1] feature setup session',
              '[testing-env/test_image:0@session1] shell: "feature2" setup session',
              '[testing-env/test_image:0/feature2@session1] feature setup session',
              '[testing-env/test_image:0] setting_up -> in_session',
              "[testing-env/test_image:0] session 'session1' started",
              '[testing-env/test_image:0] in_session -> exiting_session',
              '[testing-env/test_image:0@session1] shell: "feature1" teardown session',
              '[testing-env/test_image:0/feature1@session1] feature teardown session',
              '[testing-env/test_image:0@session1] shell: "feature2" teardown session',
              '[testing-env/test_image:0/feature2@session1] feature teardown session',
              "[testing-env/test_image:0] session 'session1' ended",
              '[testing-env/test_image:0] exiting_session -> acquired',
              '[testing-env/test_image:0] acquired -> shutting_down',
              '[testing-env/test_image:0@<idle>] shell: "feature1" teardown',
              '[testing-env/test_image:0/feature1] feature teardown',
              '[testing-env/test_image:0/feature2] feature teardown with ValueError',
              '[testing-env/test_image:0] shutting_down -> offline',
              '[testing-env/test_image:0] sandbox shutdown with FeatureTeardownError',
              # pylint: enable=line-too-long
          ]
      )

  def test_feature_teardown_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(
                simulate_teardown_error=interface.SandboxStateError
            ),
            'feature2': TestingFeature(
            ),
        },
    )
    with env:
      with env.sandbox('session1') as sb:
        pass
      self.assertEqual(len(sb.state_errors), 1)
      self.assertEqual(
          self.event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env] environment started',
              '[testing-env/test_image:0@<idle>] shell: "feature1" setup',
              '[testing-env/test_image:0/feature1] feature setup',
              '[testing-env/test_image:0@<idle>] shell: "feature2" setup',
              '[testing-env/test_image:0/feature2] feature setup',
              '[testing-env/test_image:0] created -> ready',
              '[testing-env/test_image:0] sandbox started',
              '[testing-env/test_image:0] ready -> acquired',
              '[testing-env/test_image:0] acquired -> setting_up',
              '[testing-env/test_image:0@session1] shell: "feature1" setup session',
              '[testing-env/test_image:0/feature1@session1] feature setup session',
              '[testing-env/test_image:0@session1] shell: "feature2" setup session',
              '[testing-env/test_image:0/feature2@session1] feature setup session',
              '[testing-env/test_image:0] setting_up -> in_session',
              "[testing-env/test_image:0] session 'session1' started",
              '[testing-env/test_image:0] in_session -> exiting_session',
              '[testing-env/test_image:0@session1] shell: "feature1" teardown session',
              '[testing-env/test_image:0/feature1@session1] feature teardown session',
              '[testing-env/test_image:0@session1] shell: "feature2" teardown session',
              '[testing-env/test_image:0/feature2@session1] feature teardown session',
              "[testing-env/test_image:0] session 'session1' ended",
              '[testing-env/test_image:0] exiting_session -> acquired',
              '[testing-env/test_image:0] acquired -> shutting_down',
              '[testing-env/test_image:0/feature1] feature teardown with SandboxStateError',
              '[testing-env/test_image:0@<idle>] shell: "feature2" teardown',
              '[testing-env/test_image:0/feature2] feature teardown',
              '[testing-env/test_image:0] shutting_down -> offline',
              '[testing-env/test_image:0] sandbox shutdown with FeatureTeardownError',
              # pylint: enable=line-too-long
          ]
      )

  def test_feature_setup_session_non_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(
                simulate_setup_session_error=ValueError
            ),
        },
    )
    with env:
      with self.assertRaises(ValueError):
        with env.sandbox('session1') as sb:
          sb.shell('echo "hello"')
      self.assertEqual(
          self.event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env] environment started',
              '[testing-env/test_image:0@<idle>] shell: "feature1" setup',
              '[testing-env/test_image:0/feature1] feature setup',
              '[testing-env/test_image:0@<idle>] shell: "feature2" setup',
              '[testing-env/test_image:0/feature2] feature setup',
              '[testing-env/test_image:0] created -> ready',
              '[testing-env/test_image:0] sandbox started',
              '[testing-env/test_image:0] ready -> acquired',
              '[testing-env/test_image:0] acquired -> setting_up',
              '[testing-env/test_image:0@session1] shell: "feature1" setup session',
              '[testing-env/test_image:0/feature1@session1] feature setup session',
              '[testing-env/test_image:0/feature2@session1] feature setup session with ValueError',
              "[testing-env/test_image:0] session 'session1' started with ValueError",
              '[testing-env/test_image:0] setting_up -> shutting_down',
              '[testing-env/test_image:0@session1] shell: "feature1" teardown',
              '[testing-env/test_image:0/feature1] feature teardown',
              '[testing-env/test_image:0@session1] shell: "feature2" teardown',
              '[testing-env/test_image:0/feature2] feature teardown',
              '[testing-env/test_image:0] shutting_down -> offline',
              '[testing-env/test_image:0] sandbox shutdown'
              # pylint: enable=line-too-long
          ]
      )

  def test_feature_teardown_session_non_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(
                simulate_teardown_session_error=ValueError
            ),
            'feature2': TestingFeature(),
        },
    )
    with env:
      with self.assertRaises(interface.SessionTeardownError):
        with env.sandbox('session1') as sb:
          sb.shell('echo "hello"')
      self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
      self.assertEqual(
          self.event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env] environment started',
              '[testing-env/test_image:0@<idle>] shell: "feature1" setup',
              '[testing-env/test_image:0/feature1] feature setup',
              '[testing-env/test_image:0@<idle>] shell: "feature2" setup',
              '[testing-env/test_image:0/feature2] feature setup',
              '[testing-env/test_image:0] created -> ready',
              '[testing-env/test_image:0] sandbox started',
              '[testing-env/test_image:0] ready -> acquired',
              '[testing-env/test_image:0] acquired -> setting_up',
              '[testing-env/test_image:0@session1] shell: "feature1" setup session',
              '[testing-env/test_image:0/feature1@session1] feature setup session',
              '[testing-env/test_image:0@session1] shell: "feature2" setup session',
              '[testing-env/test_image:0/feature2@session1] feature setup session',
              '[testing-env/test_image:0] setting_up -> in_session',
              "[testing-env/test_image:0] session 'session1' started",
              '[testing-env/test_image:0@session1] shell: echo "hello"',
              '[testing-env/test_image:0] in_session -> exiting_session',
              '[testing-env/test_image:0/feature1@session1] feature teardown session with ValueError',
              '[testing-env/test_image:0@session1] shell: "feature2" teardown session',
              '[testing-env/test_image:0/feature2@session1] feature teardown session',
              "[testing-env/test_image:0] session 'session1' ended",
              '[testing-env/test_image:0] exiting_session -> acquired',
              '[testing-env/test_image:0] acquired -> shutting_down',
              '[testing-env/test_image:0@<idle>] shell: "feature1" teardown',
              '[testing-env/test_image:0/feature1] feature teardown',
              '[testing-env/test_image:0@<idle>] shell: "feature2" teardown',
              '[testing-env/test_image:0/feature2] feature teardown',
              '[testing-env/test_image:0] shutting_down -> offline',
              '[testing-env/test_image:0] sandbox shutdown',
              # pylint: enable=line-too-long
          ]
      )

  def test_feature_teardown_session_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(
                simulate_teardown_session_error=interface.SandboxStateError
            ),
            'feature2': TestingFeature(),
        },
    )
    with env:
      with env.sandbox('session1') as sb:
        sb.shell('echo "hello"')
      self.assertEqual(len(sb.state_errors), 1)
      self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
      self.assertEqual(
          self.event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env] environment started',
              '[testing-env/test_image:0@<idle>] shell: "feature1" setup',
              '[testing-env/test_image:0/feature1] feature setup',
              '[testing-env/test_image:0@<idle>] shell: "feature2" setup',
              '[testing-env/test_image:0/feature2] feature setup',
              '[testing-env/test_image:0] created -> ready',
              '[testing-env/test_image:0] sandbox started',
              '[testing-env/test_image:0] ready -> acquired',
              '[testing-env/test_image:0] acquired -> setting_up',
              '[testing-env/test_image:0@session1] shell: "feature1" setup session',
              '[testing-env/test_image:0/feature1@session1] feature setup session',
              '[testing-env/test_image:0@session1] shell: "feature2" setup session',
              '[testing-env/test_image:0/feature2@session1] feature setup session',
              '[testing-env/test_image:0] setting_up -> in_session',
              "[testing-env/test_image:0] session 'session1' started",
              '[testing-env/test_image:0@session1] shell: echo "hello"',
              '[testing-env/test_image:0] in_session -> exiting_session',
              '[testing-env/test_image:0/feature1@session1] feature teardown session with SandboxStateError',
              '[testing-env/test_image:0@session1] shell: "feature2" teardown session',
              '[testing-env/test_image:0/feature2@session1] feature teardown session',
              "[testing-env/test_image:0] session 'session1' ended with SandboxStateError",
              '[testing-env/test_image:0] exiting_session -> acquired',
              '[testing-env/test_image:0] acquired -> shutting_down',
              '[testing-env/test_image:0@<idle>] shell: "feature1" teardown',
              '[testing-env/test_image:0/feature1] feature teardown',
              '[testing-env/test_image:0@<idle>] shell: "feature2" teardown',
              '[testing-env/test_image:0/feature2] feature teardown',
              '[testing-env/test_image:0] shutting_down -> offline',
              '[testing-env/test_image:0] sandbox shutdown',
              # pylint: enable=line-too-long
          ]
      )

  def test_feature_teardown_session_calling_end_session(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(
                call_end_session_on_teardown_session=True
            ),
            'feature2': TestingFeature(),
        },
    )
    with env:
      with env.sandbox('session1') as sb:
        sb.shell('echo "hello"')
      self.assertEqual(
          self.event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env] environment started',
              '[testing-env/test_image:0@<idle>] shell: "feature1" setup',
              '[testing-env/test_image:0/feature1] feature setup',
              '[testing-env/test_image:0@<idle>] shell: "feature2" setup',
              '[testing-env/test_image:0/feature2] feature setup',
              '[testing-env/test_image:0] created -> ready',
              '[testing-env/test_image:0] sandbox started',
              '[testing-env/test_image:0] ready -> acquired',
              '[testing-env/test_image:0] acquired -> setting_up',
              '[testing-env/test_image:0@session1] shell: "feature1" setup session',
              '[testing-env/test_image:0/feature1@session1] feature setup session',
              '[testing-env/test_image:0@session1] shell: "feature2" setup session',
              '[testing-env/test_image:0/feature2@session1] feature setup session',
              '[testing-env/test_image:0] setting_up -> in_session',
              "[testing-env/test_image:0] session 'session1' started",
              '[testing-env/test_image:0@session1] shell: echo "hello"',
              '[testing-env/test_image:0] in_session -> exiting_session',
              '[testing-env/test_image:0@session1] shell: "feature1" teardown session',
              '[testing-env/test_image:0/feature1@session1] feature teardown session',
              '[testing-env/test_image:0@session1] shell: "feature2" teardown session',
              '[testing-env/test_image:0/feature2@session1] feature teardown session',
              "[testing-env/test_image:0] session 'session1' ended",
              '[testing-env/test_image:0] exiting_session -> acquired',
              '[testing-env/test_image:0] acquired -> shutting_down',
              '[testing-env/test_image:0@<idle>] shell: "feature1" teardown',
              '[testing-env/test_image:0/feature1] feature teardown',
              '[testing-env/test_image:0@<idle>] shell: "feature2" teardown',
              '[testing-env/test_image:0/feature2] feature teardown',
              '[testing-env/test_image:0] shutting_down -> offline',
              '[testing-env/test_image:0] sandbox shutdown'
              # pylint: enable=line-too-long
          ]
      )

  def test_session_activity_non_state_error(self):
    env = self._create_env(
        pool_size=1,
        features={
            'feature1': TestingFeature(),
        },
    )
    with env:
      with env.sandbox('session1') as sb:
        with self.assertRaises(ValueError):
          sb.shell('echo foo', raise_error=ValueError)
        self.assertEqual(len(sb.state_errors), 0)
        sb.shell('echo bar')
        self.assertEqual(sb.status, interface.Sandbox.Status.IN_SESSION)
      sb.wait_until_not(interface.Sandbox.Status.SETTING_UP)
      self.assertEqual(sb.status, interface.Sandbox.Status.READY)
      self.assertEqual(
          self.event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env/test_image:0:0@<idle>] shell: "feature1" setup',
              '[testing-env/test_image:0:0/feature1] feature setup',
              '[testing-env/test_image:0:0@<idle>] shell: "feature1" setup session',
              '[testing-env/test_image:0:0/feature1@<idle>] feature setup session',
              '[testing-env/test_image:0:0] created -> ready',
              '[testing-env/test_image:0:0] sandbox started',
              '[testing-env] environment started',
              '[testing-env/test_image:0:0] ready -> acquired',
              '[testing-env/test_image:0:0] acquired -> setting_up',
              '[testing-env/test_image:0:0] setting_up -> in_session',
              "[testing-env/test_image:0:0] session 'session1' started",
              '[testing-env/test_image:0:0@session1] shell: echo foo with ValueError',
              '[testing-env/test_image:0:0@session1] shell: echo bar',
              '[testing-env/test_image:0:0] in_session -> exiting_session',
              '[testing-env/test_image:0:0@session1] shell: "feature1" teardown session',
              '[testing-env/test_image:0:0/feature1@session1] feature teardown session',
              "[testing-env/test_image:0:0] session 'session1' ended",
              '[testing-env/test_image:0:0] exiting_session -> setting_up',
              '[testing-env/test_image:0:0@<idle>] shell: "feature1" setup session',
              '[testing-env/test_image:0:0/feature1@<idle>] feature setup session',
              '[testing-env/test_image:0:0] setting_up -> ready',
              # pylint: enable=line-too-long
          ]
      )

  def test_session_activity_state_error(self):
    env = self._create_env(
        pool_size=1,
        features={
            'feature1': TestingFeature(),
        },
    )
    with env:
      with self.assertRaises(interface.SandboxStateError):
        with env.sandbox('session1') as sb:
          sb.shell('echo foo', raise_error=RuntimeError)
      self.assertEqual(len(sb.state_errors), 1)
      self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
      self.assertEqual(
          self.event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env/test_image:0:0@<idle>] shell: "feature1" setup',
              '[testing-env/test_image:0:0/feature1] feature setup',
              '[testing-env/test_image:0:0@<idle>] shell: "feature1" setup session',
              '[testing-env/test_image:0:0/feature1@<idle>] feature setup session',
              '[testing-env/test_image:0:0] created -> ready',
              '[testing-env/test_image:0:0] sandbox started',
              '[testing-env] environment started',
              '[testing-env/test_image:0:0] ready -> acquired',
              '[testing-env/test_image:0:0] acquired -> setting_up',
              '[testing-env/test_image:0:0] setting_up -> in_session',
              "[testing-env/test_image:0:0] session 'session1' started",
              '[testing-env/test_image:0:0@session1] shell: echo foo with RuntimeError',
              '[testing-env/test_image:0:0] in_session -> exiting_session',
              '[testing-env/test_image:0:0@session1] shell: "feature1" teardown session',
              '[testing-env/test_image:0:0/feature1@session1] feature teardown session',
              "[testing-env/test_image:0:0] session 'session1' ended with SandboxStateError",
              '[testing-env/test_image:0:0] exiting_session -> acquired',
              '[testing-env/test_image:0:0] acquired -> shutting_down',
              '[testing-env/test_image:0:0@<idle>] shell: "feature1" teardown',
              '[testing-env/test_image:0:0/feature1] feature teardown',
              '[testing-env/test_image:0:0] shutting_down -> offline',
              '[testing-env/test_image:0:0] sandbox shutdown',
              # pylint: enable=line-too-long
          ]
      )


class SandboxActivityTests(unittest.TestCase):

  def test_session_id(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=0
    )
    with env:
      with env.sandbox() as sb:
        self.assertRegex(sb.session_id, r'session-[0-9a-f]{7}')

      with env.test_feature() as test_feature:
        self.assertIsInstance(test_feature, TestingFeature)
        self.assertRegex(
            test_feature.session_id,
            r'test_feature-session-[0-9a-f]{7}'
        )

  def test_ping_error(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature(housekeep_interval=0)},
        pool_size=1,
        sandbox_keepalive_interval=0,
    )
    with env:
      with env.sandbox('session1') as sb:
        sb.rebind(
            simulate_ping_error=interface.SandboxStateError,
            skip_notification=True
        )
        sb.wait_until_next_housekeep()
        self.assertIn(sb.status, (sb.Status.SHUTTING_DOWN, sb.Status.OFFLINE))

  def test_housekeep_error(self):
    event_handler = TestingEventHandler(log_housekeep=False)
    env = TestingEnvironment(
        features={'test_feature': TestingFeature(housekeep_interval=0)},
        pool_size=1,
        housekeep_interval=1.0,
        outage_grace_period=0,
        outage_retry_interval=0.1,
        sandbox_keepalive_interval=0,
        event_handler=event_handler,
    )
    with env:
      with env.sandbox('session1') as sb:
        self.assertEqual(len(env.sandbox_pool), 1)
        self.assertEqual(sb.status, interface.Sandbox.Status.IN_SESSION)
        self.assertEqual(sb.session_id, 'session1')
        housekeep_count = sb.housekeep_counter
        sb.test_feature.rebind(
            simulate_housekeep_error=interface.SandboxStateError,
            skip_notification=True
        )
        while sb.housekeep_counter == housekeep_count or (
            sb.status == interface.Sandbox.Status.IN_SESSION
        ):
          time.sleep(0.01)
        time.sleep(1.0)
        self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
      env.wait_for_housekeeping()
    self.assertEqual(
        event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env/test_image:0:0@<idle>] shell: "test_feature" setup',
            '[testing-env/test_image:0:0/test_feature] feature setup',
            '[testing-env/test_image:0:0@<idle>] shell: "test_feature" setup session',
            '[testing-env/test_image:0:0] sandbox started',
            '[testing-env] environment started',
            "[testing-env/test_image:0:0] session 'session1' started",
            '[testing-env/test_image:0:0@session1] shell: "test_feature" teardown session',
            "[testing-env/test_image:0:0] session 'session1' ended with SandboxStateError",
            '[testing-env/test_image:0:0@<idle>] shell: "test_feature" teardown',
            '[testing-env/test_image:0:0/test_feature] feature teardown',
            '[testing-env/test_image:0:0] sandbox shutdown',
            '[testing-env/test_image:0:1@<idle>] shell: "test_feature" setup',
            '[testing-env/test_image:0:1/test_feature] feature setup',
            '[testing-env/test_image:0:1@<idle>] shell: "test_feature" setup session',
            '[testing-env/test_image:0:1] sandbox started',
            '[testing-env/test_image:0:1@<idle>] shell: "test_feature" teardown',
            '[testing-env/test_image:0:1/test_feature] feature teardown',
            '[testing-env/test_image:0:1] sandbox shutdown',
            '[testing-env] environment shutdown'
            # pylint: enable=line-too-long
        ]
    )


class SandboxServiceTests(unittest.TestCase):

  def setUp(self):
    super().setUp()
    self.maxDiff = None
    self.event_handler = TestingEventHandler()
    self.env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=0,
        outage_grace_period=0,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
        event_handler=self.event_handler,
        random_seed=1,
    )

  def test_service_call_activity_log(self):

    class CustomEventHandler(interface.EventHandler):

      def __init__(self):
        self.calls = []

      def on_sandbox_activity(
          self,
          name: str,
          sandbox: interface.Sandbox,
          session_id: str | None,
          duration: float,
          error: BaseException | None,
          **kwargs: Any):
        self.calls.append((session_id, name, kwargs))

      def on_feature_activity(
          self,
          name: str,
          feature: interface.Feature,
          session_id: str | None,
          duration: float,
          error: BaseException | None,
          **kwargs: Any):
        self.calls.append((session_id, name, kwargs))

    event_handler = CustomEventHandler()
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=0,
        event_handler=event_handler,
    )
    with env:
      with env.test_feature(session_id='session1') as test_feature:
        test_feature.call_with_varargs('sum', 1, 2, debug=True)
    self.assertEqual(
        event_handler.calls,
        [
            (None, 'shell', {'code': '"test_feature" setup'}),
            ('session1', 'shell', {'code': '"test_feature" setup session'}),
            ('session1', 'test_feature.call_with_varargs', {'args': (1, 2), 'code': 'sum', 'debug': True}),   # pylint: disable=line-too-long
            ('session1', 'shell', {'code': '"test_feature" teardown session'}),
            (None, 'shell', {'code': '"test_feature" teardown'}),
        ]
    )

  def test_service_call_from_feature(self):
    with self.env:
      with self.env.sandbox('session1') as sb:
        self.assertEqual(sb.test_feature.num_shell_calls(), 2)
        self.assertEqual(sb.test_feature.num_shell_calls(), 2)
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/test_image:0@<idle>] shell: "test_feature" setup',
            '[testing-env/test_image:0/test_feature] feature setup',
            '[testing-env/test_image:0] sandbox started',
            '[testing-env/test_image:0@session1] shell: "test_feature" setup session',
            "[testing-env/test_image:0] session 'session1' started",
            '[testing-env/test_image:0/test_feature@session1] test_feature.num_shell_calls: None',
            '[testing-env/test_image:0/test_feature@session1] test_feature.num_shell_calls: None',
            '[testing-env/test_image:0@session1] shell: "test_feature" teardown session',
            "[testing-env/test_image:0] session 'session1' ended",
            '[testing-env/test_image:0@<idle>] shell: "test_feature" teardown',
            '[testing-env/test_image:0/test_feature] feature teardown',
            '[testing-env/test_image:0] sandbox shutdown',
            '[testing-env] environment shutdown',
            # pylint: enable=line-too-long
        ]
    )

  def test_service_call_from_feature_with_error(self):
    with self.env:
      with self.assertRaises(interface.SandboxStateError):
        with self.env.sandbox('session1') as sb:
          sb.test_feature.bad_shell_call()
      self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
      self.assertEqual(len(sb.state_errors), 1)
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/test_image:0@<idle>] shell: "test_feature" setup',
            '[testing-env/test_image:0/test_feature] feature setup',
            '[testing-env/test_image:0] sandbox started',
            '[testing-env/test_image:0@session1] shell: "test_feature" setup session',
            "[testing-env/test_image:0] session 'session1' started",
            '[testing-env/test_image:0@session1] shell: bad command with RuntimeError',
            '[testing-env/test_image:0/test_feature@session1] test_feature.bad_shell_call: None with SandboxStateError',
            '[testing-env/test_image:0@session1] shell: "test_feature" teardown session',
            "[testing-env/test_image:0] session 'session1' ended with SandboxStateError",
            '[testing-env/test_image:0@<idle>] shell: "test_feature" teardown',
            '[testing-env/test_image:0/test_feature] feature teardown',
            '[testing-env/test_image:0] sandbox shutdown',
            '[testing-env] environment shutdown'
            # pylint: enable=line-too-long
        ]
    )

  def test_service_call_from_environment(self):
    with self.env:
      with self.env.test_feature() as test_feature:
        self.assertEqual(test_feature.num_shell_calls(), 2)
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/test_image:0@<idle>] shell: "test_feature" setup',
            '[testing-env/test_image:0/test_feature] feature setup',
            '[testing-env/test_image:0] sandbox started',
            '[testing-env/test_image:0@test_feature-session-2291d8c] shell: "test_feature" setup session',
            "[testing-env/test_image:0] session 'test_feature-session-2291d8c' started",
            '[testing-env/test_image:0/test_feature@test_feature-session-2291d8c] test_feature.num_shell_calls: None',
            '[testing-env/test_image:0@test_feature-session-2291d8c] shell: "test_feature" teardown session',
            "[testing-env/test_image:0] session 'test_feature-session-2291d8c' ended",
            '[testing-env/test_image:0@<idle>] shell: "test_feature" teardown',
            '[testing-env/test_image:0/test_feature] feature teardown',
            '[testing-env/test_image:0] sandbox shutdown',
            '[testing-env] environment shutdown'
            # pylint: enable=line-too-long
        ]
    )

  def test_service_call_from_environment_with_error(self):
    with self.env:
      with self.assertRaises(interface.SandboxStateError):
        with self.env.test_feature(session_id='session1') as test_feature:
          test_feature.bad_shell_call()
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/test_image:0@<idle>] shell: "test_feature" setup',
            '[testing-env/test_image:0/test_feature] feature setup',
            '[testing-env/test_image:0] sandbox started',
            '[testing-env/test_image:0@session1] shell: "test_feature" setup session',
            "[testing-env/test_image:0] session 'session1' started",
            '[testing-env/test_image:0@session1] shell: bad command with RuntimeError',
            '[testing-env/test_image:0/test_feature@session1] test_feature.bad_shell_call: None with SandboxStateError',
            '[testing-env/test_image:0@session1] shell: "test_feature" teardown session',
            "[testing-env/test_image:0] session 'session1' ended with SandboxStateError",
            '[testing-env/test_image:0@<idle>] shell: "test_feature" teardown',
            '[testing-env/test_image:0/test_feature] feature teardown',
            '[testing-env/test_image:0] sandbox shutdown',
            '[testing-env] environment shutdown',
            # pylint: enable=line-too-long
        ]
    )

  def test_service_context_manager_from_feature(self):
    with self.env:
      with self.env.sandbox('session1') as sb:
        with sb.test_feature.my_service() as service:
          service.do('hello')
        sb.shell('foo')
        self.assertEqual(sb.status, interface.Sandbox.Status.IN_SESSION)
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/test_image:0@<idle>] shell: "test_feature" setup',
            '[testing-env/test_image:0/test_feature] feature setup',
            '[testing-env/test_image:0] sandbox started',
            '[testing-env/test_image:0@session1] shell: "test_feature" setup session',
            "[testing-env/test_image:0] session 'session1' started",
            '[testing-env/test_image:0@session1] shell: hello',
            '[testing-env/test_image:0@session1] shell: foo',
            '[testing-env/test_image:0@session1] shell: "test_feature" teardown session',
            "[testing-env/test_image:0] session 'session1' ended",
            '[testing-env/test_image:0@<idle>] shell: "test_feature" teardown',
            '[testing-env/test_image:0/test_feature] feature teardown',
            '[testing-env/test_image:0] sandbox shutdown',
            '[testing-env] environment shutdown',
            # pylint: enable=line-too-long
        ]
    )

  def test_service_context_manager_from_feature_with_error(self):
    with self.env:
      with self.assertRaises(interface.SandboxStateError):
        with self.env.sandbox('session1') as sb:
          with sb.test_feature.my_service() as service:
            service.do('hello', raise_error=interface.SandboxStateError)
      self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
      self.assertEqual(len(sb.state_errors), 1)
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/test_image:0@<idle>] shell: "test_feature" setup',
            '[testing-env/test_image:0/test_feature] feature setup',
            '[testing-env/test_image:0] sandbox started',
            '[testing-env/test_image:0@session1] shell: "test_feature" setup session',
            "[testing-env/test_image:0] session 'session1' started",
            '[testing-env/test_image:0@session1] shell: hello with SandboxStateError',
            '[testing-env/test_image:0@session1] shell: "test_feature" teardown session',
            "[testing-env/test_image:0] session 'session1' ended with SandboxStateError",
            '[testing-env/test_image:0@<idle>] shell: "test_feature" teardown',
            '[testing-env/test_image:0/test_feature] feature teardown',
            '[testing-env/test_image:0] sandbox shutdown',
            '[testing-env] environment shutdown',
            # pylint: enable=line-too-long
        ]
    )

  def test_service_context_manager_from_environment(self):
    with self.env:
      with self.env.test_feature(session_id='session1') as test_feature:
        with test_feature.my_service() as service:
          service.do('foo')

      with self.env.test_feature() as test_feature:
        with test_feature.my_service() as service:
          service.do('bar')
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/test_image:0@<idle>] shell: "test_feature" setup',
            '[testing-env/test_image:0/test_feature] feature setup',
            '[testing-env/test_image:0] sandbox started',
            '[testing-env/test_image:0@session1] shell: "test_feature" setup session',
            "[testing-env/test_image:0] session 'session1' started",
            '[testing-env/test_image:0@session1] shell: foo',
            '[testing-env/test_image:0@session1] shell: "test_feature" teardown session',
            "[testing-env/test_image:0] session 'session1' ended",
            '[testing-env/test_image:0@<idle>] shell: "test_feature" teardown',
            '[testing-env/test_image:0/test_feature] feature teardown',
            '[testing-env/test_image:0] sandbox shutdown',
            '[testing-env/test_image:1@<idle>] shell: "test_feature" setup',
            '[testing-env/test_image:1/test_feature] feature setup',
            '[testing-env/test_image:1] sandbox started',
            '[testing-env/test_image:1@test_feature-session-2291d8c] shell: "test_feature" setup session',
            "[testing-env/test_image:1] session 'test_feature-session-2291d8c' started",
            '[testing-env/test_image:1@test_feature-session-2291d8c] shell: bar',
            '[testing-env/test_image:1@test_feature-session-2291d8c] shell: "test_feature" teardown session',
            "[testing-env/test_image:1] session 'test_feature-session-2291d8c' ended",
            '[testing-env/test_image:1@<idle>] shell: "test_feature" teardown',
            '[testing-env/test_image:1/test_feature] feature teardown',
            '[testing-env/test_image:1] sandbox shutdown',
            '[testing-env] environment shutdown',
            # pylint: enable=line-too-long
        ]
    )

  def test_service_context_manager_from_environment_with_error(self):
    with self.env:
      with self.assertRaises(interface.SandboxStateError):
        with self.env.test_feature() as test_feature:
          with test_feature.my_service() as service:
            service.do('hello', raise_error=interface.SandboxStateError)
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/test_image:0@<idle>] shell: "test_feature" setup',
            '[testing-env/test_image:0/test_feature] feature setup',
            '[testing-env/test_image:0] sandbox started',
            '[testing-env/test_image:0@test_feature-session-2291d8c] shell: "test_feature" setup session',
            "[testing-env/test_image:0] session 'test_feature-session-2291d8c' started",
            '[testing-env/test_image:0@test_feature-session-2291d8c] shell: hello with SandboxStateError',
            '[testing-env/test_image:0@test_feature-session-2291d8c] shell: "test_feature" teardown session',
            "[testing-env/test_image:0] session 'test_feature-session-2291d8c' ended with SandboxStateError",
            '[testing-env/test_image:0@<idle>] shell: "test_feature" teardown',
            '[testing-env/test_image:0/test_feature] feature teardown',
            '[testing-env/test_image:0] sandbox shutdown',
            '[testing-env] environment shutdown',
            # pylint: enable=line-too-long
        ]
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Schema and prompting protocols for structured data."""

# pylint: disable=g-bad-import-order
# pylint: disable=g-importing-member
from langfun.core.structured.schema.base import Schema
from langfun.core.structured.schema.base import SchemaError
from langfun.core.structured.schema.base import SchemaType

from langfun.core.structured.schema.base import class_dependencies
from langfun.core.structured.schema.base import schema_spec
from langfun.core.structured.schema.base import annotation

from langfun.core.structured.schema.base import PromptingProtocol
from langfun.core.structured.schema.base import schema_repr
from langfun.core.structured.schema.base import value_repr
from langfun.core.structured.schema.base import parse_value

from langfun.core.structured.schema.base import mark_missing
from langfun.core.structured.schema.base import Missing
from langfun.core.structured.schema.base import MISSING
from langfun.core.structured.schema.base import Unknown
from langfun.core.structured.schema.base import UNKNOWN

# JSON protocol.
from langfun.core.structured.schema.json import JsonError
from langfun.core.structured.schema.json import JsonPromptingProtocol
from langfun.core.structured.schema.json import cleanup_json

# Python protocol.
from langfun.core.structured.schema.python import PythonPromptingProtocol
from langfun.core.structured.schema.python import source_form
from langfun.core.structured.schema.python import structure_from_python

from langfun.core.structured.schema.python import class_definition
from langfun.core.structured.schema.python import class_definitions
from langfun.core.structured.schema.python import include_method_in_prompt



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class for language models through REST APIs."""

import functools
from typing import Annotated, Any, Callable

import langfun.core as lf
# Placeholder for Google-internal internet access import.
import requests


class REST(lf.LanguageModel):
  """Base class for language models accessed via REST APIs.

  The `REST` class provides a foundation for implementing language models
  that are accessed through RESTful endpoints. It handles the details of
  making HTTP requests, managing sessions, and handling common errors like
  timeouts and connection issues.

  Subclasses need to implement the `request` and `result` methods to
  convert Langfun messages to API-specific request formats and to parse
  API responses back into `LMSamplingResult` objects. They also need to
  provide the `api_endpoint` and can override `headers` for authentication.
  """

  api_endpoint: Annotated[
      str,
      'The endpoint of the REST API.'
  ]

  request: Annotated[
      Callable[[lf.Message, lf.LMSamplingOptions], dict[str, Any]],
      'A function to convert a Langfun message to a JSON request.'
  ]

  result: Annotated[
      Callable[[dict[str, Any]], lf.LMSamplingResult],
      'A function to convert a JSON response to an LMSamplingResult.'
  ]

  model: Annotated[
      str | None,
      'Model ID.'
  ] = None

  headers: Annotated[
      dict[str, Any] | None,
      'The headers for the REST API.'
  ] = None

  @functools.cached_property
  def _api_initialized(self) -> bool:
    """Returns whether the API is initialized."""
    self._initialize()
    return True

  def _initialize(self) -> None:
    """Initializes the API. Subclasses can override."""

  def session(self) -> requests.Session:
    assert self._api_initialized
    s = self._session()
    # Placeholder for Google-internal session adapter.
    s.headers.update(self.headers or {})
    return s

  def _session(self) -> requests.Session:
    """Creates a new session."""
    return requests.Session()

  def _on_bound(self):
    super()._on_bound()
    self.__dict__.pop('_api_initialized', None)

  def _sample(self, prompts: list[lf.Message]) -> list[lf.LMSamplingResult]:
    assert self._api_initialized
    return self._parallel_execute_with_currency_control(
        self._sample_single, prompts
    )

  def _sample_single(self, prompt: lf.Message) -> lf.LMSamplingResult:
    try:
      with self.session() as session:
        return self._parse_response(
            session.post(
                self.api_endpoint,
                json=self.request(prompt, self.sampling_options),
                timeout=self.timeout,
            )
        )
    except (
        requests.exceptions.Timeout,
        requests.exceptions.ReadTimeout,
        requests.exceptions.ConnectTimeout,
        TimeoutError,
    ) as e:
      raise lf.TemporaryLMError(str(e)) from e
    except (
        requests.exceptions.ConnectionError,
        requests.exceptions.ChunkedEncodingError,
        ConnectionError,
        ConnectionResetError,
    ) as e:
      error_message = str(e)
      if 'REJECTED_CLIENT_THROTTLED' in error_message:
        raise lf.TemporaryLMError(error_message) from e
      if 'UNREACHABLE_NO_RESPONSE' in error_message:
        raise lf.TemporaryLMError(error_message) from e
      if 'UNREACHABLE_ERROR' in error_message:
        raise lf.TemporaryLMError(error_message) from e
      if 'Connection reset by peer' in error_message:
        raise lf.TemporaryLMError(error_message) from e
      raise lf.LMError(error_message) from e

  def _error(self, status_code: int, content: str) -> lf.LMError:
    if status_code == 429:
      error_cls = lf.RateLimitError
    elif status_code in (
        500,  # Server side issue (might be bug).
        502,  # Bad gateway (upstream issue, might retry).
        503,  # Servers currently under load, retry after a brief wait.
        529,  # Overloaded, retry after a brief wait.
        499,  # Client Closed Request
    ):
      error_cls = lf.TemporaryLMError
    else:
      error_cls = lf.LMError
    return error_cls(f'{status_code}: {content}')

  def _parse_response(self, response: requests.Response) -> lf.LMSamplingResult:
    """Parses Anthropic's response."""
    if response.status_code == 200:
      return self.result(response.json())
    else:
      raise self._error(response.status_code, response.content)

  @property
  def max_concurrency(self) -> int | None:
    """Returns the max concurrency for this model."""
    rate_limits = self.model_info.rate_limits
    if rate_limits is not None:
      return self.estimate_max_concurrency(
          max_requests_per_minute=rate_limits.max_requests_per_minute,
          max_tokens_per_minute=rate_limits.max_tokens_per_minute
      )
    return None



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interfaces for environments, sandboxes and features."""

import abc
import contextlib
import dataclasses
import enum
import functools
import os
from typing import Annotated, Any, Callable, ContextManager, ClassVar, Iterator, Optional, Sequence, Type

import pyglove as pg

#
# Environment errors.
#


class EnvironmentError(RuntimeError):  # pylint: disable=redefined-builtin
  """Base class for environment errors."""

  def __init__(
      self,
      message: str,
      *args,
      environment: 'Environment',
      **kwargs
  ) -> None:
    self.environment = environment
    super().__init__(f'[{environment.id}] {message}.', *args, **kwargs)


class EnvironmentOutageError(EnvironmentError):
  """Error that indicates environment is offline."""

  def __init__(
      self,
      message: str | None = None,
      *args,
      offline_duration: float,
      **kwargs
  ):
    self.offline_duration = offline_duration
    super().__init__(
        message or f'Environment is offline for {offline_duration} seconds.',
        *args,
        **kwargs
    )


class EnvironmentOverloadError(EnvironmentError):
  """Error that indicates environment is overloaded."""

  def __init__(
      self,
      message: str | None = None,
      *args,
      **kwargs
  ):
    super().__init__(
        message or 'All sandboxes in the pool are either busy or dead.',
        *args, **kwargs
    )


class SandboxError(RuntimeError):
  """Base class for sandbox errors."""

  def __init__(
      self,
      message: str,
      *args,
      sandbox: 'Sandbox',
      **kwargs
  ) -> None:
    self.sandbox = sandbox
    super().__init__(f'[{sandbox.id}] {message}.', *args, **kwargs)


class SandboxStateError(SandboxError):
  """Error that indicates sandbox is in an unexpected state.

  This error is raised when the sandbox is in an unexpected state and cannot
  be recovered. As a result, the sandbox will be shutdown and user session
  will be terminated.
  """

  def __init__(
      self,
      message: str | None = None,
      *args,
      code: str | None = None,
      **kwargs
  ):
    default_message = 'Sandbox is in an unexpected state'
    if code is not None:
      default_message = (
          f'Sandbox is in an unexpected state after executing code: {code!r}'
      )
    super().__init__(message or default_message, *args, **kwargs)


class FeatureTeardownError(SandboxError):
  """Base class for feature errors."""

  def __init__(
      self,
      message: str | None = None,
      *args,
      errors: dict[str, BaseException],
      **kwargs
  ):
    self.errors = errors
    super().__init__(
        (message or
         f'Feature teardown failed with user-defined errors: {errors}.'),
        *args,
        **kwargs
    )

  @property
  def has_non_sandbox_state_error(self) -> bool:
    """Returns True if the feature teardown error has non-sandbox state error."""
    return any(
        not isinstance(e, SandboxStateError) for e in self.errors.values()
    )


class SessionTeardownError(SandboxError):
  """Base class for session errors."""

  def __init__(
      self,
      message: str | None = None,
      *args,
      errors: dict[str, BaseException],
      **kwargs
  ):
    self.errors = errors
    super().__init__(
        (message or
         f'Session teardown failed with user-defined errors: {errors}.'),
        *args,
        **kwargs
    )

  @property
  def has_non_sandbox_state_error(self) -> bool:
    """Returns True if the feature teardown error has non-sandbox state error."""
    return any(
        not isinstance(e, SandboxStateError) for e in self.errors.values()
    )

#
# Interface for sandbox-based environment.
#


class Environment(pg.Object):
  """Base class for an environment.

  An **Environment** is the central component for managing sandboxes and
  **Features**. It acts as an abstraction layer, hiding the implementation
  details of the underlying container/sandboxing system
  (e.g., Docker, virtual machines).

  The core goal is to enable the development of features that are **agnostic**
  to how or where they are executed (sandboxed or not).

  -----------------------------------------------------------------------------

  ## Core Concepts

  1.  **Sandbox-Based Features:** Features that require isolated execution
      contexts (sandboxes). They become available as properties of the sandbox
      object (e.g., `sandbox.feature1`). Applicability is determined by an image
      ID regex.
  2.  **Non-Sandbox-Based Features:** Features that run directly within the host
      process or manage their own execution context outside of the Environment's
      managed sandboxes. They are accessed directly via the Environment (e.g.,
      `env.feature3()`).

  How to Use:

  The primary usage patterns are creating a **sandbox session** or directly
  accessing a specific **Feature**, which transparently handles sandbox creation
  if needed.

  ```python
  env = MyEnvironment(
      image_ids=['image1', 'image2'],
      features={
          'feature1': Feature1(applicable_images=['image1.*']),
          'feature2': Feature2(applicable_images=['image2.*']),
          'feature3': Feature3(is_sandbox_based=False)
      })

  # Context manager for the Environment lifetime.
  with env:

    # 1. Access a specific sandbox directly.
    # Upon exiting the context, the sandbox will be shutdown or returned to
    # the pool for reuse.
    with env.sandbox(image_id='image1') as sandbox:
      # Execute a shell command inside the sandbox.
      sandbox.shell('echo "hello world"')

      # Access a sandbox-based feature (feature1 is applicable to image1).
      sandbox.feature1.feature_method()

      # Attempts to access inapplicable features will raise AttributeError:
      # sandbox.feature2  # Not applicable to image1.
      # sandbox.feature3  # Not sandbox-based.

    # 2. Access a sandbox-based feature and let the Environment manage the
    # sandbox. A suitable sandbox (e.g., one built from an image matching
    # 'image1.*') will be provisioned, and the feature instance will be yielded.
    with env.feature1() as feature1:
      feature1.feature_method()

    # 3. Access a non-sandbox-based feature.
    with env.feature3() as feature3:
      feature3.feature_method()
  ```

  -----------------------------------------------------------------------------

  ## Multi-tenancy and Pooling

  The Environment supports multi-tenancy (working with multiple image types) and
  pooling (reusing sandboxes) to amortize setup costs across different user
  requests. Pooling is configured via the pool_size parameter.

  | pool_size Value           | Behavior                                       |
  |---------------------------|------------------------------------------------|
  | 0 or (0, 0)	              | No pooling. Sandboxes are created and shut down|
  |                           | on demand (useful for local development).      |
  |---------------------------|------------------------------------------------|
  | (MIN, MAX) tuple          | Global Pool: Applies the same minimum and      |
  |                           | maximum pool size to sandboxes created from all|
  |                           | specified images.                              |
  |---------------------------|------------------------------------------------|
  | {image_regex: (MIN, MAX)} | Per-Image Pool: Allows customizing pool        |
  |                           | settings based on image ID regular expressions.|
  |                           | (e.g., 'image1.*': (64, 256), '.*': (0, 256)). |


  **Example 1: No Pooling (pool_size=0)**
  Sandboxes are created and shutdown immediately upon session end.

  ```python
  env = MyEnvironment(image_ids=['image1', 'image2'], pool_size=0)

  # Sandbox created and shutdown on demand.
  with env.sandbox(image_id='image1') as sandbox1:
    ...
  ```

  **Exaxmple 2: Global Pooling (pool_size=(0, 256))**
  Up to 256 sandboxes will be created and pooled across both images as needed.
  None are created initially.

  ```python
  env = MyEnvironment(
      image_ids=['image1', 'image2'],
      pool_size=(0, 256)
  )
  ```

  **Example 3: Per-Image Custom Pooling**:
  For images matching 'image1.*': 64 sandboxes are pre-created (MIN=64) and
  pooled, up to a MAX of 256.
  For all other images ('.*'): Sandboxes are created and pooled on demand
  (MIN=0), up to a MAX of 256.

  ```python
  env = MyEnvironment(
      image_ids=['image1', 'image2'],
      pool_size={
          'image1.*': (64, 256),
          '.*': (0, 256),
      }
  )
  ```

  ## Handling Sandbox Failures

  Sandboxes often run in distributed, ephemeral environments and must be treated
  as fault-tolerant. Langfun provides a protocol for handling unexpected sandbox
  state issues.

  ### Communicating Errors
  If a feature encounters an unexpected state in its sandbox (e.g., a process
  died), it should raise `lf.env.SandboxStateError`.

  The sandbox will be automatically shut down when its context manager exits.
  The Environment handles replacement and ensures future requests are routed to
  healthy sandboxes.

  For example:
  ```
  with env:
    with env.sandbox() as sb:
      # If SandboxStateError is raised within this block, the sandbox
      # will be forcefully shut down upon block exit.
      sb.shell('echo hi')
  ```

  ### Robust User Code
  A simple strategy for robust user code is to wrap critical operations in a
  retry loop:
  ```
  while True:
    try:
      result = do_something_that_involves_sandbox()
      break  # Success!
    except lf.env.SandboxStateError:
      # The sandbox failed; a new, healthy one will be provisioned on the next
      # iteration.
      # Wait briefly to avoid resource thrashing.
      time.sleep(1)
    except lf.env.EnvironmentOutageError:
      # If the Environment is down for too long
      # (past BaseEnvironment.outage_grace_period)
      # and cannot provision a healthy replacement, this error is raised.
      # The retry loop should be broken or an outer failure reported.
      raise
  ```
  """

  # Disable symbolic comparison and hashing for environment objects.
  use_symbolic_comparison = False

  @dataclasses.dataclass(frozen=True)
  class Id:
    """Identifier for an environment."""
    environment_id: str

    def __str__(self) -> str:
      return self.environment_id

    def working_dir(self, root_dir: str | None) -> str | None:
      """Returns the download directory for the service."""
      if root_dir is None:
        return None
      return os.path.join(root_dir, _make_path_compatible(self.environment_id))

  class Status(enum.Enum):
    """Environment state.

    State transitions:

    +---------------+               +-----------+
    |  <CREATED>    | --(start)---> |  ONLINE   | -(shutdown or outage detected)
    +---------------+      |        +-----------+              |
                           |        +-----------+              |
                           +------> | <OFFLINE> | <------------+
                                    +-----------+
    """
    CREATED = 'created'
    ONLINE = 'online'
    SHUTTING_DOWN = 'shutting_down'
    OFFLINE = 'offline'

  features: Annotated[
      dict[str, 'Feature'],
      'Features to be exposed by the environment.'
  ] = {}

  _ENV_STACK: Annotated[
      ClassVar[list['Environment']],
      'Recording the environments stacked through context managers.'
  ] = []

  #
  # Subclasses must implement:
  #

  @property
  @abc.abstractmethod
  def id(self) -> Id:
    """Returns the identifier for the environment."""

  @property
  @abc.abstractmethod
  def image_ids(self) -> list[str]:
    """Returns the non-dynamic image IDs served by the environment."""

  def image_id_for(self, feature: 'Feature') -> str:
    """Returns the default image ID for the environment."""
    for image_id in self.image_ids:
      if feature.is_applicable(image_id):
        return image_id
    raise ValueError(
        f'No image ID found for feature {feature.name} in {self.image_ids}.'
    )

  def non_sandbox_based_features(self) -> Iterator['Feature']:
    """Returns non-sandbox-based features."""
    for feature in self.features.values():
      if not feature.is_sandbox_based:
        yield feature

  @property
  @abc.abstractmethod
  def event_handler(self) -> 'EventHandler':
    """Returns the event handler for the environment."""

  @property
  @abc.abstractmethod
  def status(self) -> Status:
    """Returns the status of the environment."""

  @abc.abstractmethod
  def stats(self) -> dict[str, Any]:
    """Returns the stats of the environment."""

  @abc.abstractmethod
  def start(self) -> None:
    """Starts the environment.

    Raises:
      EnvironmentError: If the environment is not available.
    """

  @abc.abstractmethod
  def shutdown(self) -> None:
    """Shuts down the environment.

    IMPORTANT: This method shall not raise any exceptions.
    """

  @abc.abstractmethod
  def acquire(
      self,
      image_id: str | None = None,
  ) -> 'Sandbox':
    """Acquires a free sandbox from the environment.

    Args:
      image_id: The image ID to use for the sandbox. If None, it will be
        automatically determined by the environment.

    Returns:
      A free sandbox from the environment.

    Raises:
      EnvironmentOutageError: If the environment is out of service.
      EnvironmentOverloadError: If the environment is overloaded.
    """

  @abc.abstractmethod
  def new_session_id(self, feature_hint: str | None = None) -> str:
    """Generates a new session ID."""

  #
  # Environment lifecycle.
  #

  @property
  def is_online(self) -> bool:
    """Returns True if the environment is alive."""
    return self.status == self.Status.ONLINE

  def __enter__(self) -> 'Environment':
    """Enters the environment and sets it as the current environment."""
    self.start()
    Environment._ENV_STACK.append(self)
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    """Exits the environment and reset the current environment."""
    assert Environment._ENV_STACK
    Environment._ENV_STACK.pop()
    self.shutdown()

  @classmethod
  def current(cls) -> Optional['Environment']:
    """Returns the current environment."""
    if not Environment._ENV_STACK:
      return None
    return Environment._ENV_STACK[-1]

  #
  # Environment operations.
  #

  def sandbox(
      self,
      session_id: str | None = None,
      image_id: str | None = None,
  ) -> ContextManager['Sandbox']:
    """Gets a sandbox from the environment and starts a new user session."""
    return self.acquire(image_id=image_id).new_session(
        session_id or self.new_session_id()
    )

  def __getattr__(self, name: str) -> Any:
    """Gets a feature session from a free sandbox from the environment.

    Example:
      ```
      with XboxEnvironment(
          features={'selenium': SeleniumFeature()}
      ) as env:
        with env.selenium() as selenium:
          driver = selenium.get_driver()
      ```

    Args:
      name: The name of the feature.

    Returns:
      A callable `(image_id, *, session_id) -> ContextManager[Feature]` that
      creates a context manager for the requested feature under a new client
      session.
    """
    if name in self.features:
      return _feature_session_creator(self, self.features[name])
    raise AttributeError(name)


@contextlib.contextmanager
def _sandbox_session_for_feature(
    environment: Environment,
    feature: 'Feature',
    image_id: str | None = None,
    session_id: str | None = None,
) -> Iterator['Feature']:
  """Returns a context manager for a session for a feature."""
  assert feature.is_sandbox_based
  if image_id is None:
    image_id = environment.image_id_for(feature)
  elif not feature.is_applicable(image_id):
    raise ValueError(
        f'Feature {feature.name!r} is not applicable to image {image_id!r}.'
    )
  sandbox = environment.acquire(image_id=image_id)
  with sandbox.new_session(
      session_id=session_id or environment.new_session_id(feature.name)
  ):
    yield sandbox.features[feature.name]


def _feature_session_creator(environment: Environment, feature: 'Feature'):
  """Returns a callable that returns a context manager for a feature session."""
  def fn(session_id: str | None = None, image_id: str | None = None):
    if feature.is_sandbox_based:
      return _sandbox_session_for_feature(
          environment, feature, image_id, session_id
      )
    assert image_id is None, (
        'Non-sandbox based feature does not support image ID.'
    )
    return feature.new_session(
        session_id or environment.new_session_id(feature.name)
    )
  return fn


# Enable automatic conversion from str to Environment.Id.
pg.typing.register_converter(str, Environment.Id, Environment.Id)


class Sandbox(pg.Object):
  """Interface for sandboxes.

  A sandbox is a container that runs a single image with a set of features.
  It will be brought up by the environment, setup the features, fullfill user
  requests, and then tear down features and finally the sandbox itself.
  """

  # Disable symbolic comparison and hashing for sandbox objects.
  use_symbolic_comparison = False

  @dataclasses.dataclass(frozen=True, slots=True)
  class Id:
    """Identifier for a sandbox."""
    environment_id: Environment.Id
    image_id: str
    sandbox_id: str

    def __str__(self) -> str:
      return f'{self.environment_id}/{self.image_id}:{self.sandbox_id}'

    def working_dir(self, root_dir: str | None) -> str | None:
      """Returns the download directory for the sandbox."""
      if root_dir is None:
        return None
      return os.path.join(
          self.environment_id.working_dir(root_dir),
          _make_path_compatible(self.image_id),
          _make_path_compatible(self.sandbox_id)
      )

  class Status(enum.Enum):
    r"""Sandbox state.

    State transitions:

                    (sandbox / feature
    +------------+   teardown)            +---------------+
    | <OFFLINE>  | <--------------------- | SHUTTING_DOWN |
    +------------+                        +---------------+
                                            ^     ^
                                           /       \
                          (setup failed)  /         \
                                         /           \
    +-----------+   (start)  +------------+           \
    | <CREATED> | -------->  | SETTING_UP |            \
    +-----------+         ^  +------------+             \
                         /        |                      \
                        /         | (sandbox /            \
                       /          |  feature /session      \
                      /           v  setup succeeded)       \
                     /        +---------+                    \
                    /         |  READY  |                     \
                   /          +---------+                      \
                  /                |                            \
                 /                 |  (acquire)                  \
                /                  v                              \
               /              +----------+                         \
              |               | ACQUIRED |                          \
              |               +----------+                           |
              |                    |                                 |
              |                    |  (start_session)                |
              |               +------------+                         |
              |               | SETTING_UP |-- (setup failed) ------>+
              |               +------------+                         |
              |                    |                                 |
              |                    v  (succeeded)                    |
              |               +--------------+                       |
              |               |  IN_SESSION  |- (op failed) -------->+
              |               +--------------+                       |
              |                    |                                 |
              |                    |  (end_session)                  |
              |                    |                                 |
              |                    v             (session teardown   |
      (setup next            +-----------------+  failed OR          |
      session for  <---------| EXITING_SESSION |- non-reusable  -----+
      reusable sandbox)      +-----------------+  sandbox)

    """

    # The sandbox is created, but not yet started.
    CREATED = 'created'

    # The sandbox is being setting up to serve user sessions.
    SETTING_UP = 'setting_up'

    # The sandbox is set up and free to be acquired by the user.
    READY = 'ready'

    # The sandbox is acquired by a thread, but not yet in a user session.
    ACQUIRED = 'acquired'

    # The sandbox is in a user session.
    IN_SESSION = 'in_session'

    # The sandbox is exiting a user session.
    EXITING_SESSION = 'exiting_session'

    # The sandbox is being shut down.
    SHUTTING_DOWN = 'shutting_down'

    # The sandbox is offline.
    OFFLINE = 'offline'

    @property
    def is_online(self) -> bool:
      """Returns True if the sandbox is online."""
      return self in (
          Sandbox.Status.SETTING_UP,
          Sandbox.Status.READY,
          Sandbox.Status.ACQUIRED,
          Sandbox.Status.IN_SESSION,
      )

  @property
  @abc.abstractmethod
  def id(self) -> Id:
    """Returns the identifier for the sandbox."""

  @property
  @abc.abstractmethod
  def image_id(self) -> str:
    """Returns the image ID used for bootstrapping the sandbox."""

  @property
  @abc.abstractmethod
  def environment(self) -> Environment:
    """Returns the environment for the sandbox."""

  @property
  @abc.abstractmethod
  def features(self) -> dict[str, 'Feature']:
    """Returns the features in the sandbox."""

  @property
  @abc.abstractmethod
  def status(self) -> Status:
    """Returns the status of the sandbox."""

  @property
  def is_online(self) -> bool:
    """Returns True if the sandbox is online."""
    return self.status.is_online

  @abc.abstractmethod
  def report_state_error(self, error: SandboxStateError) -> None:
    """Reports state error the sandbox.

    If state errors are reported, the sandbox will be forcefully shutdown when
    `Sandbox.end_session()` is called, even if the sandbox is set to be
    reusable.

    Args:
      error: SandboxStateError to report.
    """

  @abc.abstractmethod
  def start(self) -> None:
    """Starts the sandbox.

    State transitions:
      CREATED -> SETTING_UP -> READY: When all sandbox and feature setup
        succeeds.
      CREATED -> SETTING_UP -> SHUTTING_DOWN -> OFFLINE: When sandbox or feature
        setup fails.

    `start` and `shutdown` should be called in pairs, even when the sandbox
    fails to start. This ensures proper cleanup.

    Start may fail with two sources of errors:

    1. SandboxStateError: If sandbox or feature setup fail due to enviroment
       outage or sandbox state errors.
    2. BaseException: If feature setup failed with user-defined errors, this
       could happen when there is bug in the user code or non-environment code
       failure.

    In both cases, the sandbox will be shutdown automatically, and the error
    will be add to `errors`. The sandbox is considered dead and will not be
    further used.

    Raises:
      SandboxStateError: If the sandbox is in a bad state.
      BaseException: If feature setup failed with user-defined errors.
    """

  @abc.abstractmethod
  def shutdown(self) -> None:
    """Shuts down the sandbox.

    State transitions:
      SHUTTING_DOWN -> SHUTTING_DOWN: No operation.
      OFFLINE -> OFFLINE: No operation.
      SETTING_UP -> SHUTTING_DOWN -> OFFLINE: When sandbox and feature
        setup fails.
      IN_SESSION -> SHUTTING_DOWN -> OFFLINE: When user session exits while
        sandbox is set not to reuse, or session teardown fails.
      FREE -> SHUTTING_DOWN -> OFFLINE: When sandbox is shutdown when the
        environment is shutting down, or housekeeping loop shuts down the
        sandbox due to housekeeping failures.


    Please be aware that `shutdown` will be called whenever an operation on the
    sandbox encounters a critical error. This means, `shutdown` should not make
    the assumption that the sandbox is in a healthy state, even `start` could
    fail. As a result, `shutdown` must allow re-entry and be thread-safe with
    other sandbox operations.

    Shutdown may fail with two sources of errors:

    1. SandboxStateError: If the sandbox is in a bad state, and feature teardown
       logic depending on a healthy sandbox may fail. In such case, we do not
       raise error to the user as the user session is considered completed. The
       sandbox is abandoned and new user sessions will be served on other
       sandboxes.

    2. BaseException: The sandbox is in good state, but user code raises error
       due to bug or non-environment code failure. In such case, errors will be
       raised to the user so the error could be surfaced and handled properly.
       The sandbox is treated as shutdown and will not be further used.

    Raises:
      BaseException: If feature teardown failed with user-defined errors.
    """

  @abc.abstractmethod
  def start_session(
      self,
      session_id: str,
  ) -> None:
    """Begins a user session with the sandbox.

    State transitions:
      ACQUIRED -> SETTING_UP -> IN_SESSION: When session setup succeeds.
      ACQUIRED -> SETTING_UP -> SHUTTING_DOWN -> OFFLINE: When session setup
        fails.

    A session is a sequence of stateful interactions with the sandbox.
    Across different sessions the sandbox are considered stateless.
    `start_session` and `end_session` should always be called in pairs, even
    when the session fails to start. `Sandbox.new_session` context manager is
    the recommended way to use `start_session` and `end_session` in pairs.

    Starting a session may fail with two sources of errors:

    1. SandboxStateError: If the sandbox is in a bad state or session setup
       failed.

    2. BaseException: If session setup failed with user-defined errors.

    In both cases, the sandbox will be shutdown automatically and the
    session will be considered ended. The error will be added to `errors`.
    Future session will be served on other sandboxes.

    Args:
      session_id: The identifier for the user session.

    Raises:
      SandboxStateError: If the sandbox is already in a bad state or session
        setup failed.
      BaseException: If session setup failed with user-defined errors.
    """

  @abc.abstractmethod
  def end_session(self) -> None:
    """Ends the user session with the sandbox.

    State transitions:
      IN_SESSION -> EXITING_SESSION -> READY: When user session exits normally,
        and sandbox is set to reuse.
      IN_SESSION -> EXITING_SESSION -> SHUTTING_DOWN -> OFFLINE: When user
        session exits while
        sandbox is set not to reuse, or session teardown fails.
      IN_SESSION -> EXITING_SESSION -> SETTING_UP -> READY: When user session
        exits normally, and sandbox is set to reuse, and proactive session setup
        is enabled.
      IN_SESSION -> EXITING_SESSION -> SETTING_UP -> SHUTTING_DOWN -> OFFLINE:
        When user session exits normally, and proactive session setup is enabled
        but fails.
      EXITING_SESSION -> EXITING_SESSION: No operation.
      not IN_SESSION -> same state: No operation

    `end_session` should always be called for each `start_session` call, even
    when the session fails to start, to ensure proper cleanup.

    When `end_session` is called with state errors reported, the sandbox will be
    forcefully shutdown even if the sandbox is set to be reusable.

    `end_session` may fail with two sources of errors:

    1. SandboxStateError: If the sandbox is in a bad state or session teardown
       failed.

    2. BaseException: If session teardown failed with user-defined errors.

    In both cases, the sandbox will be shutdown automatically and the
    session will be considered ended. The error will be added to `errors`.
    Future session will be served on other sandboxes.

    However, SandboxStateError encountered during `end_session` will NOT be
    raised to the user as the user session is considered completed.

    Raises:
      BaseException: If session teardown failed with user-defined errors.
    """

  @property
  @abc.abstractmethod
  def session_id(self) -> str | None:
    """Returns the current user session identifier."""

  @abc.abstractmethod
  def track_activity(
      self,
      name: str,
      **kwargs: Any
  ) -> ContextManager[None]:
    """Context manager that tracks a sandbox activity.

    Args:
      name: The name of the activity.
      **kwargs: Additional keyword arguments to pass to the activity handler.

    Returns:
      A context manager that tracks the activity, including duration and error.
    """

  #
  # API related to a user session.
  # A sandbox could be reused across different user sessions.
  # A user session is a sequence of stateful interactions with the sandbox,
  # Across different sessions the sandbox are considered stateless.
  #

  @contextlib.contextmanager
  def new_session(self, session_id: str) -> Iterator['Sandbox']:
    """Context manager for obtaining a sandbox for a user session.

    State transitions:
      ACQUIRED -> IN_SESSION -> READY: When session setup and teardown succeed.
      ACQUIRED -> IN_SESSINO -> OFFLINE: When session setup or teardown fails.

    Args:
      session_id: The identifier for the user session.

    Yields:
      The sandbox for the user session.

    Raises:
      SandboxStateError: If a session cannot be started on the sandbox.
      BaseException: If session setup or teardown failed with user-defined
        errors.
    """
    self.start_session(session_id)
    try:
      yield self
    except SandboxStateError as e:
      self.report_state_error(e)
      raise
    finally:
      self.end_session()

  def __getattr__(self, name: str) -> Any:
    """Gets a feature from current sandbox.

    Example:
      ```
      with MyEnvironment(
          features={'feature1': Feature1()}
      ) as env:
        with env.sandbox('session1') as sb:
          driver = sb.feature1.feature_method()
      ```

    Args:
      name: The name of the feature.

    Returns:
      A feature from current sandbox.
    """
    if name in self.features:
      return self.features[name]
    raise AttributeError(name)


class Feature(pg.Object):
  """Interface for features that run in a Langfun environment.

  There are two type of features: sandbox-based and non-sandbox-based.
  Sandbox-based features run in a sandbox, which is emulated in a separate
  process. Non-sandbox-based features do not run in a sandbox.

  Features can be directly accessed through the environment, for example:

  ```python
  env = MyEnvironment(
      feature={
          'feature1': SandboxBasedFeature(),
          'feature2': NonSandboxBasedFeature(),
      }
  )
  # Start the environment.
  with env:
    # Access feature1, which involves acquiring a sandbox and return the feature
    # associated with the sandbox.
    with env.feature1() as f1:
      f1.feature_method()

    # Access feature2, which does not involve acquiring a sandbox.
    with env.feature2() as f2:
      f2.feature_method()
  ```

  Sandbox-based features can also be accessed through the sandbox, for example:

  ```python
  with env.sandbox('session1') as sb:
    # Access feature1 within the sandbox.
    sb.feature1.feature_method()

    # Attribute error.
    sb.feature2
  ```
  """

  @dataclasses.dataclass
  class Id:
    container_id: Environment.Id | Sandbox.Id
    feature_name: str

    def __str__(self) -> str:
      return f'{self.container_id}/{self.feature_name}'

    def working_dir(self, root_dir: str | None) -> str | None:
      """Returns the working directory for the feature."""
      if root_dir is None:
        return None
      return os.path.join(
          self.container_id.working_dir(root_dir),
          _make_path_compatible(self.feature_name)
      )

  # Disable symbolic comparison and hashing for sandbox objects.
  allow_symbolic_comparison = False

  @property
  @abc.abstractmethod
  def name(self) -> str:
    """Name of the feature, which will be used as key to access the feature."""

  @functools.cached_property
  def id(self) -> Id:
    """Returns the identifier of the feature."""
    if self.is_sandbox_based:
      return Feature.Id(self.sandbox.id, self.name)
    return Feature.Id(self.environment.id, self.name)

  @property
  @abc.abstractmethod
  def environment(self) -> Environment:
    """Returns the environment that the feature is running in."""

  @property
  @abc.abstractmethod
  def is_sandbox_based(self) -> bool:
    """Returns True if the feature is sandbox-based."""

  @property
  @abc.abstractmethod
  def sandbox(self) -> Sandbox | None:
    """Returns the sandbox that the feature is running in.

    Returns:
      The sandbox that the feature is running in. None if the feature is not
      sandbox-based or not yet bound with a sandbox.
    """

  @abc.abstractmethod
  def is_applicable(self, image_id: str) -> bool:
    """Returns True if the feature is applicable to the given image."""

  @abc.abstractmethod
  def setup(self, sandbox: Sandbox | None = None) -> None:
    """Sets up the feature.

    For sandbox-based features, the setup will be called when a sandbox is
    started for the first time.

    For non-sandbox-based features, the setup will be called when the
    environment starts.

    When a feature's `setup` is called, its `teardown` is guaranteed to be
    called.

    State transitions:
      SETTING_UP -> READY: When setup succeeds.
      SETTING_UP -> SHUTTING_DOWN -> OFFLINE: When setup fails.

    `setup` is called when a sandbox is started for the first time. When a
    feature's `setup` is called, its `teardown` is guaranteed to be called.

    Args:
      sandbox: The sandbox that the feature is running in.

    Raises:
      SandboxStateError: If setup failed due to sandbox state errors.
      BaseException: If setup failed with user-defined errors.
    """

  @abc.abstractmethod
  def teardown(self) -> None:
    """Teardowns the feature, which is called once when the sandbox is down.

    State transitions:
      SHUTTING_DOWN -> OFFLINE: When teardown succeeds or fails.

    Raises:
      SandboxStateError: If the sandbox is in a bad state.
    """

  @abc.abstractmethod
  def setup_session(self) -> None:
    """Sets up the feature for the upcoming user session.

    State transitions:
      SETTING_UP -> READY: When session setup succeeds.
      SETTING_UP -> SHUTTING_DOWN -> OFFLINE: When sessino setup fails.

    `setup_session` is called when a new user session starts for per-session
    setup. `setup_session` and `teardown_session` will be called in pairs, even
    `setup_session` fails.

    `setup_session` may fail with two sources of errors:

    1. SandboxStateError: If the sandbox is in a bad state or session setup
       failed.
    2. BaseException: If session setup failed with user-defined errors.

    In both cases, the error will be raised to the user and the session will be
    ended. The sandbox will shutdown automatically and will not be further used.

    Raises:
      SandboxStateError: If the sandbox is in a bad state or session setup
        failed.
      BaseException: If session setup failed with user-defined errors.
    """

  @abc.abstractmethod
  def teardown_session(self) -> None:
    """Teardowns the feature for an ending user session.

    State transitions:
      SHUTTING_DOWN -> OFFLINE: When session teardown succeeds or fails.

    `teardown_session` is called when a user session ends for per-
    session teardown. `teardown_session` will always be called upon a feature
    whose `setup_session` is called.

    `teardown_session` may fail with two sources of errors:

    1. SandboxStateError: If the sandbox is in a bad state or session teardown
       failed.
    2. BaseException: If session teardown failed with user-defined errors.

    In both cases, the session will be closed and the sandbox will be shutdown.
    However, SandboxStateError encountered during `teardown_session` will NOT be
    raised to the user as the user session is considered completed. Other errors
    will be raised to the user for proper error handling.

    Raises:
      BaseException: If session teardown failed with user-defined errors.
    """

  @abc.abstractmethod
  def housekeep(self) -> None:
    """Performs housekeeping for the feature.

    State transitions:
      (no state change): When housekeeping succeeds.
      original state -> SHUTTING_DOWN -> OFFLINE: When housekeeping fails.

    Raises:
      SandboxStateError: If the sandbox is in a bad state.
    """

  @property
  @abc.abstractmethod
  def housekeep_interval(self) -> int | None:
    """Returns the interval in seconds for feature housekeeping.

    Returns:
      The interval in seconds for feature housekeeping. If None, the feature
      will not be housekeeping.
    """

  @abc.abstractmethod
  def track_activity(
      self,
      name: str,
      **kwargs: Any
  ) -> ContextManager[None]:
    """Context manager that tracks a feature activity.

    Args:
      name: The name of the activity.
      **kwargs: Additional keyword arguments to pass to the activity handler.

    Returns:
      A context manager that tracks the activity, including duration and error.
    """

  @property
  def session_id(self) -> str | None:
    """Returns the current user session identifier."""
    if self.is_sandbox_based:
      return self.sandbox.session_id
    return self._non_sandbox_based_session_id

  @contextlib.contextmanager
  def new_session(self, session_id: str) -> Iterator['Feature']:
    """Context manager for obtaining a non-sandbox-based feature session."""
    assert not self.is_sandbox_based, (
        'Applicable only to non-sandbox-based features. '
        'For sandbox-based features, use `Sandbox.new_session` instead.'
    )
    try:
      self._non_sandbox_based_session_id = session_id
      self.setup_session()
      yield self
    finally:
      try:
        # Since the session is ended, we don't want to raise any errors during
        # session teardown to the user. So we catch all exceptions here.
        # However the event handler will still be notified and log the error.
        self.teardown_session()
      except BaseException:  # pylint: disable=broad-except
        pass
      self._non_sandbox_based_session_id = None

  def _on_bound(self) -> None:
    """Called when the feature is bound to a sandbox."""
    super()._on_bound()
    self._non_sandbox_based_session_id = None


def _make_path_compatible(id_str: str) -> str:
  """Makes a path compatible with CNS."""
  return id_str.translate(
      str.maketrans({
          '@': '_',
          ':': '_',
          '#': '_',
          ' ': '',
          '<': '',
          '>': '',
      })
  )


def treat_as_sandbox_state_error(
    errors: Sequence[
        Type[BaseException] | tuple[Type[BaseException], str]
    ] | None = None
) -> Callable[..., Any]:
  """Decorator for Sandbox/Feature methods to convert errors to SandboxStateError.

  Args:
    errors: A sequence of exception types or tuples of (error_type, msg_regex).
      when matched, treat the error as SandboxStateError, which will lead to
      a sandbox shutdown when caught by `Sandbox.new_session()` context manager.

  Returns:
    The decorator function.
  """

  def decorator(func):
    @functools.wraps(func)
    def method_wrapper(self, *args, **kwargs) -> Any:
      """Helper function to safely execute logics in the sandbox."""

      assert isinstance(self, (Sandbox, Feature)), self
      sandbox = self.sandbox if isinstance(self, Feature) else self

      try:
        # Execute the service function.
        return func(self, *args, **kwargs)
      except BaseException as e:
        if pg.match_error(e, errors):
          state_error = SandboxStateError(
              'Sandbox encountered an unexpected error executing '
              f'`{func.__name__}` (args={args!r}, kwargs={kwargs!r}): {e}',
              sandbox=sandbox
          )
          raise state_error from e
        raise
    return method_wrapper
  return decorator


def log_activity(name: str | None = None):
  """Decorator for Sandbox/Feature methods to log sandbox/feature activity."""

  def decorator(func):
    signature = pg.typing.get_signature(func)
    def to_kwargs(*args, **kwargs):
      num_non_self_args = len(signature.arg_names) - 1
      if len(args) > num_non_self_args:
        assert signature.varargs is not None, (signature, args)
        kwargs[signature.varargs.name] = tuple(args[num_non_self_args:])
        args = args[:num_non_self_args]
      for i in range(len(args)):
        # The first argument is `self`.
        kwargs[signature.arg_names[i + 1]] = args[i]
      return kwargs

    @functools.wraps(func)
    def method_wrapper(self, *args, **kwargs) -> Any:
      """Helper function to safely execute logics in the sandbox."""

      assert isinstance(self, (Sandbox, Feature)), self
      with self.track_activity(
          name or func.__name__,
          **to_kwargs(*args, **kwargs)
      ):
        return func(self, *args, **kwargs)
    return method_wrapper
  return decorator


#
# Interface for event handlers.
#


class _EnvironmentEventHandler:
  """Base class for event handlers of an environment."""

  def on_environment_starting(self, environment: Environment) -> None:
    """Called when the environment is getting started.

    Args:
      environment: The environment.
    """

  def on_environment_start(
      self,
      environment: Environment,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when the environment is started.

    Args:
      environment: The environment.
      duration: The environment start duration in seconds.
      error: The error that failed the environment start. If None, the
        environment started normally.
    """

  def on_environment_housekeep(
      self,
      environment: Environment,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when the environment finishes a round of housekeeping.

    Args:
      environment: The environment.
      counter: Zero-based counter of the housekeeping round.
      duration: The environment start duration in seconds.
      error: The error that failed the housekeeping. If None, the
        housekeeping succeeded.
      **kwargs: Environment-specific properties computed during housekeeping.
    """

  def on_environment_shutting_down(
      self,
      environment: Environment,
      offline_duration: float,
  ) -> None:
    """Called when the environment is shutting down.

    Args:
      environment: The environment.
      offline_duration: The environment offline duration in seconds.
    """

  def on_environment_shutdown(
      self,
      environment: Environment,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    """Called when the environment is shutdown.

    Args:
      environment: The environment.
      duration: The environment shutdown duration in seconds.
      lifetime: The environment lifetime in seconds.
      error: The error that caused the environment to shutdown. If None, the
        environment shutdown normally.
    """


class _SandboxEventHandler:
  """Base class for sandbox event handlers."""

  def on_sandbox_start(
      self,
      sandbox: Sandbox,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox is started.

    Args:
      sandbox: The sandbox.
      duration: The time spent on starting the sandbox.
      error: The error that caused the sandbox to start. If None, the sandbox
        started normally.
    """

  def on_sandbox_status_change(
      self,
      sandbox: Sandbox,
      old_status: 'Sandbox.Status',
      new_status: 'Sandbox.Status',
      span: float,
  ) -> None:
    """Called when a sandbox status changes.

    Args:
      sandbox: The sandbox.
      old_status: The old sandbox status.
      new_status: The new sandbox status.
      span: Time spent on the old status in seconds.
    """

  def on_sandbox_shutdown(
      self,
      sandbox: Sandbox,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox is shutdown.

    Args:
      sandbox: The sandbox.
      duration: The time spent on shutting down the sandbox.
      lifetime: The sandbox lifetime in seconds.
      error: The error that caused the sandbox to shutdown. If None, the
        sandbox shutdown normally.
    """

  def on_sandbox_session_start(
      self,
      sandbox: Sandbox,
      session_id: str,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox session starts.

    Args:
      sandbox: The sandbox.
      session_id: The session ID.
      duration: The time spent on starting the session.
      error: The error that caused the session to start. If None, the session
        started normally.
    """

  def on_sandbox_session_end(
      self,
      sandbox: Sandbox,
      session_id: str,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox session ends.

    Args:
      sandbox: The sandbox.
      session_id: The session ID.
      duration: The time spent on ending the session.
      lifetime: The session lifetime in seconds.
      error: The error that caused the session to end. If None, the session
        ended normally.
    """

  def on_sandbox_activity(
      self,
      name: str,
      sandbox: Sandbox,
      session_id: str | None,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a sandbox activity is performed.

    Args:
      name: The name of the sandbox activity.
      sandbox: The sandbox.
      session_id: The session ID.
      duration: The sandbox activity duration in seconds.
      error: The error that caused the sandbox activity to perform. If None,
        the sandbox activity performed normally.
      **kwargs: The keyword arguments of the sandbox activity.
    """

  def on_sandbox_housekeep(
      self,
      sandbox: Sandbox,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a sandbox finishes a round of housekeeping.

    Args:
      sandbox: The sandbox.
      counter: Zero-based counter of the housekeeping round.
      duration: The sandbox housekeeping duration in seconds.
      error: The error that caused the sandbox to housekeeping. If None, the
        sandbox housekeeping normally.
      **kwargs: Sandbox-specific properties computed during housekeeping.
    """


class _FeatureEventHandler:
  """Base class for feature event handlers."""

  def on_feature_setup(
      self,
      feature: Feature,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is setup.

    Applicable to both sandbox-based and non-sandbox-based features.

    Args:
      feature: The feature.
      duration: The feature setup duration in seconds.
      error: The error happened during the feature setup. If None,
        the feature setup performed normally.
    """

  def on_feature_teardown(
      self,
      feature: Feature,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is teardown.

    Applicable to both sandbox-based and non-sandbox-based features.

    Args:
      feature: The feature.
      duration: The feature teardown duration in seconds.
      error: The error happened during the feature teardown. If None,
        the feature teardown performed normally.
    """

  def on_feature_teardown_session(
      self,
      feature: Feature,
      session_id: str,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a feature is teardown with a session.

    Applicable to both sandbox-based and non-sandbox-based features.

    Args:
      feature: The feature.
      session_id: The session ID.
      duration: The feature teardown session duration in seconds.
      error: The error happened during the feature teardown session. If
        None, the feature teardown session performed normally.
    """

  def on_feature_setup_session(
      self,
      feature: Feature,
      session_id: str | None,
      duration: float,
      error: BaseException | None,
  ) -> None:
    """Called when a feature is setup with a session.

    Applicable to both sandbox-based and non-sandbox-based features.

    Args:
      feature: The feature.
      session_id: The session ID.
      duration: The feature setup session duration in seconds.
      error: The error happened during the feature setup session. If
        None, the feature setup session performed normally.
    """

  def on_feature_activity(
      self,
      name: str,
      feature: Feature,
      session_id: str | None,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a feature activity is performed.

    Applicable to both sandbox-based and non-sandbox-based features.

    Args:
      name: The name of the feature activity.
      feature: The feature.
      session_id: The session ID. Session ID could be None if a feature
        activity is performed when setting up a session
        (e.g. BaseEnvironment.proactive_session_setup is on)
      duration: The feature activity duration in seconds.
      error: The error happened during the feature activity. If None,
        the feature activity performed normally.
      **kwargs: The keyword arguments of the feature activity.
    """

  def on_feature_housekeep(
      self,
      feature: Feature,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs,
  ) -> None:
    """Called when a sandbox feature is housekeeping.

    Applicable to both sandbox-based and non-sandbox-based features.

    Args:
      feature: The feature.
      counter: Zero-based counter of the housekeeping round.
      duration: The feature housekeeping duration in seconds.
      error: The error happened during the feature housekeeping. If None, the
        feature housekeeping normally.
      **kwargs: Feature-specific properties computed during housekeeping.
    """


class EventHandler(
    _EnvironmentEventHandler,
    _SandboxEventHandler,
    _FeatureEventHandler,
):
  """Base class for langfun/env handlers."""



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for MCP client."""

import unittest
from langfun.core import async_support
from langfun.core import mcp as lf_mcp
from langfun.core import message as lf_message
from mcp.server import fastmcp as fastmcp_lib

mcp = fastmcp_lib.FastMCP(host='0.0.0.0', port=1234)


@mcp.tool()
async def add(a: int, b: int) -> int:
  """Adds two integers and returns their sum.

  Args:
    a: The first integer.
    b: The second integer.

  Returns:
    The sum of the two integers.
  """
  return a + b


class McpTest(unittest.TestCase):

  def test_sync_usages(self):
    client = lf_mcp.McpClient.from_fastmcp(mcp)
    tools = client.list_tools()
    self.assertEqual(len(tools), 1)
    with client.session() as session:
      self.assertEqual(
          # Test `session.call_tool` method as `tool.__call__` is already tested
          # in `tool_test.py`.
          session.call_tool(tools['add'](a=1, b=2)), 3
      )

  def test_async_usages(self):
    async def _test():
      client = lf_mcp.McpClient.from_fastmcp(mcp)
      tools = client.list_tools()
      self.assertEqual(len(tools), 1)
      tool_cls = tools['add']
      self.assertEqual(tool_cls.__name__, 'Add')
      self.assertEqual(tool_cls.TOOL_NAME, 'add')
      async with client.session() as session:
        self.assertEqual(
            # Test `session.acall_tool` method as `tool.acall` is already
            # tested in `tool_test.py`.
            await session.acall_tool(tool_cls(a=1, b=2), returns_message=True),
            lf_message.ToolMessage(text='3', result=3)
        )
    async_support.invoke_sync(_test)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for MCP session."""

import unittest
from unittest import mock

from langfun.core.mcp import session as mcp_session
import mcp
from mcp.client import sse
from mcp.client import streamable_http


class McpSessionTest(unittest.TestCase):

  @mock.patch.object(mcp, 'stdio_client', autospec=True)
  def test_from_command(self, mock_stdio_client):
    mcp_session.McpSession.from_command('my-command', ['--foo'])
    mock_stdio_client.assert_called_once_with(
        mcp.StdioServerParameters(command='my-command', args=['--foo'])
    )

  @mock.patch.object(streamable_http, 'streamablehttp_client', autospec=True)
  def test_from_url_mcp(self, mock_streamablehttp_client):
    mcp_session.McpSession.from_url(
        'http://localhost/mcp', headers={'k': 'v'}
    )
    mock_streamablehttp_client.assert_called_once_with(
        'http://localhost/mcp', {'k': 'v'}
    )

  @mock.patch.object(sse, 'sse_client', autospec=True)
  def test_from_url_sse(self, mock_sse_client):
    mcp_session.McpSession.from_url('http://localhost/sse', headers={'k': 'v'})
    mock_sse_client.assert_called_once_with('http://localhost/sse', {'k': 'v'})

  def test_from_url_unsupported(self):
    with self.assertRaisesRegex(ValueError, 'Unsupported transport: foo'):
      mcp_session.McpSession.from_url('http://localhost/foo')


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for beam runner."""

import os
import tempfile
from typing import Any
import unittest

from langfun.core.eval.v2 import checkpointing  # pylint: disable=unused-import
from langfun.core.eval.v2 import eval_test_helper
from langfun.core.eval.v2 import reporting  # pylint: disable=unused-import
from langfun.core.eval.v2.runners import beam  # pylint: disable=unused-import
import pyglove as pg


@unittest.skip(
    'These tests are flaky due to writing ckpt files with standard IO.'
    'We will move to `beam.io` and re-enable these tests later.'
)
class BeamRunnerTest(unittest.TestCase):

  def assert_same_list(self, actual: list[Any], expected: list[Any]):
    self.assertEqual(len(actual), len(expected))
    for i, (x, y) in enumerate(zip(actual, expected)):
      if x is not y:
        print(i, pg.diff(x, y))
      self.assertIs(x, y)

  def setUp(self):
    super().setUp()
    self.test_dir = os.path.join(tempfile.mkdtemp(), 'test_dir')

  def test_basic(self):
    plugin = eval_test_helper.TestPlugin()
    exp = eval_test_helper.test_experiment()
    root_dir = os.path.join(self.test_dir, 'test_beam_runner')
    run = exp.run(
        root_dir,
        runner='beam',
        plugins=[
            plugin,
            reporting.ExampleHtmlGenerator(),
            checkpointing.PerExampleCheckpointer(
                checkpoint_filename='checkpoint.jsonl'
            ),
        ],
        concurrent_startup_delay=(1, 2),
        use_cache='no',
        ckpt_format='jsonl',
    )

    self.assertIsNotNone(plugin.start_time)
    self.assertIsNotNone(plugin.complete_time)
    self.assertGreater(plugin.complete_time, plugin.start_time)

    self.assertEqual(len(plugin.started_experiments), len(exp.nodes))
    self.assertEqual(len(plugin.completed_experiments), len(exp.nodes))
    self.assertEqual(len(plugin.started_example_ids), 6 * 10)
    self.assertEqual(len(plugin.completed_example_ids), 6 * 10)
    self.assert_same_list(plugin.skipped_experiments, [])

    for node in exp.leaf_nodes:
      for i in range(node.num_examples):
        self.assertTrue(
            pg.io.path_exists(
                run.output_path_for(node, f'{i + 1}.html')
            )
        )

    for node in exp.nodes:
      if node.is_leaf:
        self.assertTrue(node.progress.is_started)
        self.assertTrue(node.progress.is_completed)
        self.assertEqual(node.progress.num_skipped, 0)
        self.assertEqual(node.progress.num_completed, 10)
        self.assertEqual(node.progress.num_failed, 1)
      else:
        self.assertEqual(node.progress.num_skipped, 0)
        self.assertEqual(node.progress.num_failed, 0)
        self.assertEqual(node.progress.num_processed, node.progress.num_total)

    # Test warm start.
    root_dir2 = os.path.join(self.test_dir, 'test_warm_start')
    exp = eval_test_helper.test_experiment()
    plugin = eval_test_helper.TestPlugin()
    run2 = exp.run(
        root_dir2,
        warm_start_from=run.output_root,
        runner='beam',
        plugins=[plugin],
        use_cache='no',
        ckpt_format='jsonl',
    )
    for node in run2.experiment.nodes:
      if node.is_leaf:
        self.assertTrue(node.progress.is_started)
        self.assertTrue(node.progress.is_completed)
        self.assertEqual(node.progress.num_skipped, 0)
        self.assertEqual(node.progress.num_completed, 10)
        self.assertEqual(node.progress.num_failed, 1)
      else:
        self.assertEqual(node.progress.num_skipped, 0)
        self.assertEqual(node.progress.num_failed, 0)
        self.assertEqual(node.progress.num_processed, node.progress.num_total)

  def test_shuffle_inputs(self):
    root_dir = os.path.join(self.test_dir, 'test_shuffle_inputs')
    exp = eval_test_helper.test_experiment()
    plugin = eval_test_helper.TestPlugin()
    run = exp.run(
        root_dir,
        runner='beam',
        plugins=[plugin],
        shuffle_inputs=True,
        use_cache='no',
        ckpt_format='jsonl',
    )
    self.assertTrue(run.shuffle_inputs)

  def test_beam_runner_does_not_support_cache(self):
    exp = eval_test_helper.test_experiment()
    root_dir = os.path.join(self.test_dir, 'test_beam_runner_cache')
    with self.assertRaisesRegex(ValueError, 'Cache is not supported'):
      exp.run(
          root_dir,
          runner='beam',
          use_cache='global',
      )

  def test_no_beam(self):
    orig_beam = beam.beam
    beam.beam = None
    with self.assertRaisesRegex(ValueError, 'Beam is not installed'):
      exp = eval_test_helper.TestEvaluation()
      root_dir = os.path.join(self.test_dir, 'test_no_beam')
      exp.run(root_dir, runner='beam')
    beam.beam = orig_beam


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Config saver test."""

import os
import tempfile
import unittest
from langfun.core.eval.v2 import config_saver
from langfun.core.eval.v2 import eval_test_helper
from langfun.core.eval.v2.runners import parallel  # pylint: disable=unused-import


class RunConfigSaverTest(unittest.TestCase):

  def test_save_run_config(self):
    root_dir = os.path.join(tempfile.mkdtemp(), 'test_run_config_saver')
    experiment = eval_test_helper.test_evaluation()
    run = experiment.run(
        root_dir, 'new', plugins=[config_saver.RunConfigSaver()]
    )
    self.assertTrue(os.path.exists(os.path.join(run.output_root, 'run.json')))


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for MCP tool."""

import base64
import inspect
import unittest

from langfun.core import async_support
from langfun.core import message as lf_message
from langfun.core import modalities as lf_modalities
from langfun.core.mcp import client as mcp_client
from langfun.core.mcp import tool as mcp_tool
import mcp
from mcp.server import fastmcp as fastmcp_lib
import pyglove as pg


# MCP server setup for testing.
_mcp_server = fastmcp_lib.FastMCP(host='0.0.0.0', port=1235)


@_mcp_server.tool()
async def add(a: int, b: int) -> int:
  """Adds two integers."""
  return a + b


class McpToolTest(unittest.TestCase):

  def setUp(self):
    super().setUp()
    self.client = mcp_client.McpClient.from_fastmcp(_mcp_server)
    self.tools = self.client.list_tools()

  def test_snake_to_camel(self):
    self.assertEqual(mcp_tool._snake_to_camel('foo_bar'), 'FooBar')
    self.assertEqual(mcp_tool._snake_to_camel('foo'), 'Foo')

  def test_base64_decode(self):
    self.assertEqual(
        mcp_tool._base64_decode(base64.b64encode(b'foo').decode('utf-8')),
        b'foo'
    )

  def test_make_input_class(self):
    schema = pg.Schema(
        description='Foo input.',
        fields=[
            pg.typing.Field('x', pg.typing.Int(), 'Integer x.'),
            pg.typing.Field('y', pg.typing.Str(), 'String y.'),
        ],
    )
    input_cls = mcp_tool.McpToolInput.make_class('foo_input', schema)
    self.assertTrue(issubclass(input_cls, mcp_tool.McpToolInput))
    self.assertEqual(input_cls.__name__, 'FooInput')
    self.assertEqual(input_cls.__doc__, 'Foo input.')
    s = input_cls.__schema__
    self.assertEqual(list(s.fields.keys()), ['x', 'y'])
    self.assertEqual(repr(input_cls), "<input-class 'FooInput'>")
    self.assertEqual(
        repr(input_cls(x=1, y='abc')),
        "FooInput(x=1, y='abc')",
    )

  def test_make_tool_class(self):
    tool_def = mcp.Tool(
        name='my_tool',
        inputSchema={
            'type': 'object',
            'properties': {
                'a': {'type': 'integer', 'description': 'Integer a.'},
                'b': {'type': 'string', 'description': 'String b.'},
            },
            'required': ['a'],
        },
        description='My tool.',
    )
    tool_cls = mcp_tool.McpTool.make_class(tool_def)
    self.assertTrue(issubclass(tool_cls, mcp_tool.McpTool))
    self.assertEqual(tool_cls.__name__, 'MyTool')
    self.assertEqual(tool_cls.TOOL_NAME, 'my_tool')
    self.assertEqual(tool_cls.__doc__, 'My tool.')
    s = tool_cls.__schema__
    self.assertEqual(list(s.fields.keys()), ['a', 'b'])
    self.assertEqual(repr(tool_cls), "<tool-class 'MyTool'>")
    self.assertEqual(s.fields['a'].description, 'Integer a.')
    self.assertEqual(s.fields['b'].description, 'String b.')

    self.assertEqual(
        tool_cls.python_definition(markdown=True),
        inspect.cleandoc(
            """
            MyTool

            ```python
            class MyTool:
              \"\"\"My tool.\"\"\"
              # Integer a.
              a: int
              # String b.
              b: str | None
            ```
            """
        ),
    )
    self.assertEqual(
        tool_cls.python_definition(markdown=False),
        inspect.cleandoc(
            """
            MyTool

            class MyTool:
              \"\"\"My tool.\"\"\"
              # Integer a.
              a: int
              # String b.
              b: str | None
            """
        ),
    )

  def test_input_parameters(self):
    tool_cls = self.tools['add']
    self.assertEqual(tool_cls(a=1, b=2).input_parameters(), {'a': 1, 'b': 2})

  def test_result_to_message(self):
    img_data = base64.b64encode(b'image-data').decode('utf-8')
    audio_data = base64.b64encode(b'audio-data').decode('utf-8')

    tool_def = self.tools['add']
    result = mcp.types.CallToolResult(
        content=[
            mcp.types.TextContent(type='text', text='hello'),
            mcp.types.ImageContent(
                type='image', data=img_data, mimeType='image/png'
            ),
            mcp.types.AudioContent(
                type='audio', data=audio_data, mimeType='audio/wav'
            ),
        ],
        structuredContent={'x': 1},
    )
    message = tool_def.result_to_message(result)
    self.assertIsInstance(message, lf_message.ToolMessage)
    self.assertIn('hello', message.text)
    self.assertIn('<<[[image', message.text)
    self.assertIn('<<[[audio', message.text)
    self.assertEqual(message.metadata, {'x': 1})
    modalities = message.modalities()
    self.assertEqual(len(modalities), 2)
    self.assertIsInstance(modalities[0], lf_modalities.Image)
    self.assertEqual(modalities[0].to_bytes(), b'image-data')
    self.assertIsInstance(modalities[1], lf_modalities.Audio)
    self.assertEqual(modalities[1].to_bytes(), b'audio-data')

  def test_sync_call(self):
    add_tool_cls = self.tools['add']
    with self.client.session() as session:
      # Test returning structured content.
      self.assertEqual(add_tool_cls(a=1, b=2)(session), 3)

      # Test returning message.
      self.assertEqual(
          add_tool_cls(a=1, b=2)(session, returns_message=True),
          lf_message.ToolMessage(text='3', result=3),
      )

  def test_async_call(self):
    async def _test():
      add_tool_cls = self.tools['add']
      async with self.client.session() as session:
        # Test returning structured content.
        self.assertEqual(await add_tool_cls(a=1, b=2).acall(session), 3)

        # Test returning message.
        self.assertEqual(
            await add_tool_cls(a=1, b=2).acall(session, returns_message=True),
            lf_message.ToolMessage(text='3', result=3),
        )

    async_support.invoke_sync(_test)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test utils for base environment."""

import contextlib
import time
from typing import Iterator, Type

from langfun.env import base_environment
from langfun.env import base_feature
from langfun.env import base_sandbox
from langfun.env import interface
import pyglove as pg


class TestingEnvironment(base_environment.BaseEnvironment):
  """Testing environment for unit tests."""
  image_ids: list[str] = ['test_image']
  housekeep_interval: float = 0.0
  simulate_start_error: Type[BaseException] | None = None
  simulate_shutdown_error: Type[BaseException] | None = None
  simulate_ping_error: Type[BaseException] | None = None
  offline: bool = False

  __test__ = False

  @property
  def id(self) -> interface.Environment.Id:
    return interface.Environment.Id('testing-env')

  def wait_for_housekeeping(self):
    housekeep_counter = self.housekeep_counter
    while self.housekeep_counter == housekeep_counter:
      time.sleep(0.01)

  def _create_sandbox(
      self,
      image_id: str,
      sandbox_id: str,
      reusable: bool,
      proactive_session_setup: bool,
      keepalive_interval: float | None,
  ) -> base_sandbox.BaseSandbox:
    return TestingSandbox(
        environment=self,
        id=interface.Sandbox.Id(
            environment_id=self.id,
            image_id=image_id,
            sandbox_id=sandbox_id
        ),
        image_id=image_id,
        reusable=reusable,
        proactive_session_setup=proactive_session_setup,
        keepalive_interval=keepalive_interval,
        simulate_start_error=self.simulate_start_error,
        simulate_shutdown_error=self.simulate_shutdown_error,
        simulate_ping_error=self.simulate_ping_error,
    )


class TestingSandbox(base_sandbox.BaseSandbox):
  """Testing sandbox for unit tests."""

  simulate_start_error: Type[BaseException] | None = None
  simulate_shutdown_error: Type[BaseException] | None = None
  simulate_ping_error: Type[BaseException] | None = None

  __test__ = False

  def _on_bound(self) -> None:
    super()._on_bound()
    self._shell_history = []
    self._ping_history = []

  def _raise_error(self, message, error_type: Type[BaseException], **kwargs):
    if (error_type is interface.SandboxStateError or
        issubclass(error_type, interface.SandboxStateError)):
      kwargs['sandbox'] = self
      raise error_type(message, **kwargs)
    else:
      raise error_type(message)

  def wait_until_not(
      self,
      status: interface.Sandbox.Status | tuple[interface.Sandbox.Status, ...]
  ) -> None:
    if not isinstance(status, tuple):
      status = (status,)
    while self.status in status:
      time.sleep(0.01)

  def wait_until_next_housekeep(self) -> None:
    housekeep_counter = self.housekeep_counter
    while self.housekeep_counter == housekeep_counter:
      time.sleep(0.01)

  def _start(self) -> None:
    if self.simulate_start_error:
      self._raise_error('Sandbox start error', self.simulate_start_error)
    super()._start()

  def _shutdown(self) -> None:
    if self.simulate_shutdown_error:
      self._raise_error('Sandbox shutdown error', self.simulate_shutdown_error)
    super()._shutdown()

  @interface.treat_as_sandbox_state_error(errors=(RuntimeError,))
  @interface.log_activity()
  def shell(
      self,
      code: str,
      raise_error: Type[BaseException] | None = None,
  ) -> str:
    self._shell_history.append(code)
    if raise_error is not None:
      self._raise_error(f'shell "{code}" failed', raise_error)
    return f'shell "{code}" succeeded'

  def _ping(self) -> None:
    self._ping_history.append(not self.simulate_ping_error)
    if self.simulate_ping_error:
      self._raise_error('Ping error', self.simulate_ping_error, code='ping')


class TestingFeature(base_feature.BaseFeature):
  """Testing feature for unit tests."""

  housekeep_interval = 0
  setup_session_delay: float = 0.0
  simulate_housekeep_error: Type[BaseException] | None = None
  simulate_setup_error: Type[BaseException] | None = None
  simulate_teardown_error: Type[BaseException] | None = None
  simulate_setup_session_error: Type[BaseException] | None = None
  simulate_teardown_session_error: Type[BaseException] | None = None
  call_end_session_on_teardown_session: bool = False

  __test__ = False

  class Service:
    """Sandbox."""

    def __init__(self, sandbox: interface.Sandbox):
      self._sandbox = sandbox

    def do(self, code: str, raise_error: Type[BaseException] | None = None):
      self._sandbox.shell(code, raise_error=raise_error)

  def _raise_error(self, message, error_type: Type[BaseException], **kwargs):
    self._sandbox._raise_error(message, error_type, **kwargs)  # pylint: disable=protected-access

  def _setup(self) -> None:
    if self.simulate_setup_error:
      self._raise_error(f'{self.name} setup error', self.simulate_setup_error)
    self.sandbox.shell(f'"{self.name}" setup')

  def _teardown(self) -> None:
    if self.simulate_teardown_error:
      self._raise_error(
          f'{self.name} teardown error', self.simulate_teardown_error
      )
    self.sandbox.shell(f'"{self.name}" teardown')

  def _setup_session(self) -> None:
    if self.setup_session_delay > 0:
      time.sleep(self.setup_session_delay)

    if self.simulate_setup_session_error:
      self._raise_error(
          'Feature session setup error', self.simulate_setup_session_error
      )
    self.sandbox.shell(f'"{self.name}" setup session')

  def _teardown_session(self) -> None:
    if self.simulate_teardown_session_error:
      self._raise_error(
          'Feature session teardown error', self.simulate_teardown_session_error
      )
    self.sandbox.shell(f'"{self.name}" teardown session')
    if self.call_end_session_on_teardown_session:
      self.sandbox.end_session()

  @interface.log_activity()
  def num_shell_calls(self) -> int:
    return len(self.sandbox._shell_history)  # pylint: disable=protected-access

  @interface.log_activity()
  def bad_shell_call(self) -> None:
    self.sandbox.shell('bad command', raise_error=RuntimeError)

  @interface.log_activity()
  def show_session_id(self):
    return self.session_id

  @interface.log_activity()
  def call_with_varargs(self, code: str, *args, **kwargs):
    del code, args, kwargs
    return 0

  def _on_bound(self) -> None:
    super()._on_bound()
    self._service = None

  @contextlib.contextmanager
  def my_service(self) -> Iterator[Service]:
    try:
      self._service = TestingFeature.Service(sandbox=self.sandbox)
      yield self._service
    finally:
      self._service = None

  def _housekeep(self) -> None:
    if self.simulate_housekeep_error:
      raise interface.SandboxStateError(
          'House keeping error', sandbox=self.sandbox
      )


class TestingNonSandboxBasedFeature(base_feature.BaseFeature):
  """Testing non-sandbox based feature for unit tests."""
  is_sandbox_based: bool = False
  simulate_setup_error: Type[BaseException] | None = None
  simulate_teardown_error: Type[BaseException] | None = None
  simulate_setup_session_error: Type[BaseException] | None = None
  simulate_teardown_session_error: Type[BaseException] | None = None
  simulate_housekeep_error: Type[BaseException] | None = None

  __test__ = False

  def _setup(self) -> None:
    if self.simulate_setup_error:
      raise self.simulate_setup_error('Feature setup error')

  def _teardown(self) -> None:
    if self.simulate_teardown_error:
      raise self.simulate_teardown_error('Feature teardown error')

  def _setup_session(self) -> None:
    if self.simulate_setup_session_error:
      raise self.simulate_setup_session_error('Feature session setup error')

  def _teardown_session(self) -> None:
    if self.simulate_teardown_session_error:
      raise self.simulate_teardown_session_error(
          'Feature session teardown error'
      )

  def _housekeep(self) -> None:
    if self.simulate_housekeep_error:
      raise self.simulate_housekeep_error('Feature housekeeping error')
    _ = self.foo(1)

  @interface.log_activity()
  def foo(self, x: int) -> int:
    return x + 1


class TestingEventHandler(pg.Object, interface.EventHandler):
  """Testing environment event handler for unit tests."""

  log_sandbox_status: bool = False
  log_feature_setup: bool = True
  log_session_setup: bool = False
  log_housekeep: bool = False

  __test__ = False

  def _on_bound(self) -> None:
    super()._on_bound()
    self._logs = []

  @property
  def logs(self) -> list[str]:
    return self._logs

  def _add_message(self, message: str, error: BaseException | None) -> None:
    """Adds a message to the history."""
    if error is None:
      self._logs.append(message)
    else:
      self._logs.append(f'{message} with {error.__class__.__name__}')

  def on_environment_start(
      self,
      environment: interface.Environment,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when the environment is started."""
    assert duration > 0
    self._add_message(f'[{environment.id}] environment started', error)

  def on_environment_housekeep(
      self,
      environment: interface.Environment,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when the environment finishes a round of housekeeping."""
    assert duration > 0
    if self.log_housekeep:
      self._add_message(
          f'[{environment.id}] environment housekeeping {counter}', error
      )

  def on_environment_shutdown(
      self,
      environment: interface.Environment,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    """Called when the environment is shutdown."""
    assert duration > 0 and lifetime is not None
    self._add_message(f'[{environment.id}] environment shutdown', error)

  def on_sandbox_start(
      self,
      sandbox: interface.Sandbox,
      duration: float,
      error: BaseException | None
  ) -> None:
    assert duration > 0
    self._add_message(f'[{sandbox.id}] sandbox started', error)

  def on_sandbox_status_change(
      self,
      sandbox: interface.Sandbox,
      old_status: interface.Sandbox.Status,
      new_status: interface.Sandbox.Status,
      span: float
  ) -> None:
    assert span > 0
    if self.log_sandbox_status:
      self._add_message(
          f'[{sandbox.id}] {old_status.value} -> {new_status.value}',
          None,
      )

  def on_sandbox_shutdown(
      self,
      sandbox: interface.Sandbox,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    assert duration > 0 and lifetime is not None
    self._add_message(f'[{sandbox.id}] sandbox shutdown', error)

  def on_sandbox_session_start(
      self,
      sandbox: interface.Sandbox,
      session_id: str,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox session starts."""
    assert duration > 0
    self._add_message(
        f'[{sandbox.id}] session {session_id!r} started', error
    )

  def on_sandbox_session_end(
      self,
      sandbox: interface.Sandbox,
      session_id: str,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox session ends."""
    assert duration > 0 and lifetime > 0
    self._add_message(
        f'[{sandbox.id}] session {session_id!r} ended', error
    )

  def on_sandbox_activity(
      self,
      name: str,
      sandbox: interface.Sandbox,
      session_id: str | None,
      duration: float,
      error: BaseException | None,
      *,
      code: str | None = None,
      **kwargs
  ) -> None:
    """Called when a sandbox activity is performed."""
    del kwargs
    log_id = f'{sandbox.id}@{session_id or "<idle>"}'
    self._add_message(
        f'[{log_id}] {name}: {code}', error
    )

  def on_sandbox_housekeep(
      self,
      sandbox: interface.Sandbox,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    assert duration > 0
    if self.log_housekeep:
      self._add_message(
          f'[{sandbox.id}] sandbox housekeeping {counter}', error
      )

  def on_feature_setup(
      self,
      feature: interface.Feature,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is setup."""
    assert duration > 0
    if self.log_feature_setup:
      self._add_message(
          f'[{feature.id}] feature setup', error
      )

  def on_feature_teardown(
      self,
      feature: interface.Feature,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is teardown."""
    assert duration > 0
    if self.log_feature_setup:
      self._add_message(
          f'[{feature.id}] feature teardown', error
      )

  def on_feature_setup_session(
      self,
      feature: interface.Feature,
      session_id: str | None,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is setup."""
    assert duration > 0
    if self.log_session_setup:
      self._add_message(
          f'[{feature.id}@{session_id or "<idle>"}] feature setup session',
          error
      )

  def on_feature_teardown_session(
      self,
      feature: interface.Feature,
      session_id: str,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is teardown."""
    assert duration > 0
    if self.log_session_setup:
      self._add_message(
          f'[{feature.id}@{session_id}] feature teardown session', error
      )

  def on_feature_activity(
      self,
      name: str,
      feature: interface.Feature,
      session_id: str | None,
      duration: float,
      error: BaseException | None,
      *,
      code: str | None = None,
      **kwargs
  ) -> None:
    """Called when a sandbox activity is performed."""
    del kwargs
    log_id = f'{feature.id}@{session_id or "<idle>"}'
    self._add_message(
        f'[{log_id}] {name}: {code}', error
    )

  def on_feature_housekeep(
      self,
      feature: interface.Feature,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a sandbox feature is housekeeping."""
    assert duration > 0
    if self.log_housekeep:
      self._add_message(
          f'[{feature.id}] feature housekeeping {counter}', error
      )



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Environment for LLM agents."""

# pylint: disable=g-importing-member, g-bad-import-order, g-import-not-at-top
from langfun.env.interface import EnvironmentError   # pylint: disable=redefined-builtin
from langfun.env.interface import EnvironmentOutageError
from langfun.env.interface import EnvironmentOverloadError
from langfun.env.interface import SandboxError
from langfun.env.interface import SandboxStateError

from langfun.env.interface import Environment
from langfun.env.interface import Sandbox
from langfun.env.interface import Feature

from langfun.env.interface import EventHandler

# Decorators for sandbox/feature methods.
from langfun.env.interface import treat_as_sandbox_state_error
from langfun.env.interface import log_activity

from langfun.env.base_environment import BaseEnvironment
from langfun.env.base_sandbox import BaseSandbox
from langfun.env.base_feature import BaseFeature

from langfun.env import load_balancers
from langfun.env.load_balancers import LoadBalancer

# Import all event handlers.
from langfun.env.event_handlers import *

# Google-internal imports.
